"""Custom scoring functions"""

import numpy as np

from skorch.dataset import unpack_data


def loss_scoring(net, X, y=None, sample_weight=None):
    """Calculate score using the criterion of the net

    Use the exact same logic as during model training to calculate the score.

    This function can be used to implement the ``score`` method for a
    :class:`.NeuralNet` through sub-classing. This is useful, for example, when
    combining skorch models with sklearn objects that rely on the model's
    ``score`` method. For example:

    >>> class ScoredNet(skorch.NeuralNetClassifier):
    ...     def score(self, X, y=None):
    ...         return loss_scoring(self, X, y)

    Parameters
    ----------
    net : skorch.NeuralNet
        A fitted Skorch :class:`.NeuralNet` object.

    X : input data, compatible with skorch.dataset.Dataset
        By default, you should be able to pass:

          * numpy arrays
          * torch tensors
          * pandas DataFrame or Series
          * scipy sparse CSR matrices
          * a dictionary of the former three
          * a list/tuple of the former three
          * a Dataset

        If this doesn't work with your data, you have to pass a
        ``Dataset`` that can deal with the data.

    y : target data, compatible with skorch.dataset.Dataset
        The same data types as for ``X`` are supported. If your X is a Dataset
        that contains the target, ``y`` may be set to None.

    sample_weight : array-like of shape (n_samples,)
        Sample weights.

    Returns
    -------
    loss_value : float32 or np.ndarray
        Return type depends on ``net.criterion_.reduction``, and will be a float
        if reduction is ``'sum'`` or ``'mean'``. If reduction is ``'none'`` then
        this function returns a ``np.ndarray`` object.

    """
    if sample_weight is not None:
        raise NotImplementedError(
            "sample_weight for loss_scoring is not yet supported."
        )

    net.check_is_fitted()

    dataset = net.get_dataset(X, y)
    iterator = net.get_iterator(dataset, training=False)
    history = {"loss": [], "batch_size": []}
    reduction = net.criterion_.reduction
    if reduction not in ["mean", "sum", "none"]:
        raise ValueError(
            "Expected one of 'mean', 'sum' or 'none' "
            "for reduction but got {reduction}.".format(reduction=reduction)
        )

    for batch in iterator:
        yp = net.evaluation_step(batch, training=False)
        yi = unpack_data(batch)[1]
        loss = net.get_loss(yp, yi)
        if reduction == "none":
            loss_value = loss.detach().cpu().numpy()
        else:
            loss_value = loss.item()
        history["loss"].append(loss_value)
        history["batch_size"].append(yi.size(0))

    if reduction == "none":
        return np.concatenate(history["loss"], 0)
    if reduction == "sum":
        return np.sum(history["loss"])
    return np.average(history["loss"], weights=history["batch_size"])



"""Tests for dataset.py."""

import unittest
from unittest.mock import Mock

import numpy as np
import pytest
from scipy import sparse
from sklearn.datasets import make_classification
import torch
import torch.utils.data
from torch import nn
import torch.nn.functional as F
from torch.utils.data import DataLoader

from skorch.utils import data_from_dataset
from skorch.utils import is_torch_data_type
from skorch.utils import to_tensor
from skorch.tests.conftest import pandas_installed


class TestGetLen:
    @pytest.fixture
    def get_len(self):
        from skorch.dataset import get_len
        return get_len

    @pytest.mark.parametrize('data, expected', [
        (np.zeros(5), 5),
        (np.zeros((3, 4, 5)), 3),
        ([np.zeros(5), np.zeros((5, 4))], 5),
        ((np.zeros((5, 4)), np.zeros(5)), 5),
        ({'0': np.zeros(3), '1': np.zeros((3, 4))}, 3),
        (torch.zeros(5), 5),
        (torch.zeros((3, 4, 5)), 3),
        ([torch.zeros(5), torch.zeros((5, 4))], 5),
        ((torch.zeros((5, 4)), torch.zeros(5)), 5),
        (sparse.csr_matrix(np.zeros((5, 3))), 5),
        ({'0': torch.zeros(3), '1': torch.zeros((3, 4))}, 3),
        ([0, 1, 2], 3),
        ([[0, 1, 2], [3, 4, 5]], 3),
        ({'0': [0, 1, 2], '1': (3, 4, 5)}, 3),
        ((
            [0, 1, 2],
            np.zeros(3),
            torch.zeros(3),
            sparse.csr_matrix(np.zeros((3, 5))),
            {'0': (1, 2, 3)}),
         3),
    ])
    def test_valid_lengths(self, get_len, data, expected):
        length = get_len(data)
        assert length == expected

    @pytest.mark.parametrize('data', [
        [np.zeros(5), np.zeros((4, 5))],
        {'0': np.zeros(3), '1': np.zeros((4, 3))},
        [torch.zeros(5), torch.zeros((4, 5))],
        {'0': torch.zeros(3), '1': torch.zeros((4, 3))},
        [[0, 1, 2], [3, 4]],
        ([0, 1, 2], [3, 4]),
        {'0': [0, 1, 2], '1': (3, 4)},
        ([0, 1, 2], np.zeros(3), torch.zeros(2), {'0': (1, 2, 3)}),
    ])
    def test_inconsistent_lengths(self, get_len, data):
        with pytest.raises(ValueError):
            get_len(data)

    def test_get_len_transformers_tokenizer(self, get_len):
        transformers = pytest.importorskip('transformers')

        X = ['hello there'] * 10
        tokenizer = transformers.AutoTokenizer.from_pretrained('bert-base-uncased')
        tokens = tokenizer(X)
        assert get_len(tokens) == 10


class TestNetWithoutY:
    net_fixture_params = [
        {'classification': True, 'batch_size': 1},
        {'classification': False, 'batch_size': 1},
        {'classification': True, 'batch_size': 2},
        {'classification': False, 'batch_size': 2},
    ]

    @pytest.fixture
    def net_cls_1d(self):
        from skorch.toy import make_regressor
        return make_regressor(
            input_units=1,
            num_hidden=0,
            output_units=1,
        )

    @pytest.fixture
    def net_cls_2d(self):
        from skorch.toy import make_regressor
        return make_regressor(
            input_units=2,
            num_hidden=0,
            output_units=1,
        )

    @pytest.fixture
    def loader_clf(self):
        class Loader(torch.utils.data.DataLoader):
            def __iter__(self):
                z = super().__iter__()
                return ((x, torch.zeros(x.size(0)).long()) for x, _ in z)
        return Loader

    @pytest.fixture
    def loader_reg(self):
        class Loader(torch.utils.data.DataLoader):
            def __iter__(self):
                z = super().__iter__()
                return ((x, torch.zeros(x.size(0), 1).float()) for x, _ in z)
        return Loader

    @pytest.fixture
    def train_split(self):
        from skorch.dataset import ValidSplit
        return ValidSplit(0.2, stratified=False)

    @pytest.fixture(params=net_fixture_params)
    def net_1d(self, request, net_cls_1d, train_split):
        if request.param['classification']:
            from skorch import NeuralNetClassifier
            wrap_cls = NeuralNetClassifier
        else:
            from skorch import NeuralNetRegressor
            wrap_cls = NeuralNetRegressor

        return wrap_cls(
            net_cls_1d,
            max_epochs=2,
            train_split=train_split,
            batch_size=request.param['batch_size']
        )

    @pytest.fixture(params=net_fixture_params)
    def net_2d(self, request, net_cls_2d, train_split):
        if request.param['classification']:
            from skorch import NeuralNetClassifier
            wrap_cls = NeuralNetClassifier
        else:
            from skorch import NeuralNetRegressor
            wrap_cls = NeuralNetRegressor

        return wrap_cls(
            net_cls_2d,
            max_epochs=2,
            train_split=train_split,
            batch_size=request.param['batch_size']
        )

    @pytest.fixture(params=net_fixture_params)
    def net_1d_custom_loader(self, request, net_cls_1d,
                             loader_clf, loader_reg, train_split):
        """Parametrized fixture returning a NeuralNet
        classifier/regressor, for different batch sizes, working on 1d
        data.

        """
        if request.param['classification']:
            from skorch import NeuralNetClassifier
            wrap_cls = NeuralNetClassifier
            loader = loader_clf
        else:
            from skorch import NeuralNetRegressor
            wrap_cls = NeuralNetRegressor
            loader = loader_reg

        return wrap_cls(
            net_cls_1d,
            iterator_train=loader,
            iterator_valid=loader,
            max_epochs=2,
            train_split=train_split,
            batch_size=request.param['batch_size']
        )

    @pytest.fixture(params=net_fixture_params)
    def net_2d_custom_loader(self, request, net_cls_2d,
                             loader_clf, loader_reg, train_split):
        """Parametrized fixture returning a NeuralNet
        classifier/regressor, for different batch sizes, working on 2d
        data.

        """
        if request.param['classification']:
            from skorch import NeuralNetClassifier
            wrap_cls = NeuralNetClassifier
            loader = loader_clf
        else:
            from skorch import NeuralNetRegressor
            wrap_cls = NeuralNetRegressor
            loader = loader_reg

        return wrap_cls(
            net_cls_2d,
            iterator_train=loader,
            iterator_valid=loader,
            max_epochs=2,
            train_split=train_split,
            batch_size=request.param['batch_size']
        )

    def test_net_1d_tensor_raises_error(self, net_1d):
        X = torch.arange(0, 8).view(-1, 1).long()
        # We expect check_data to throw an exception
        # because we did not specify a custom data loader.
        with pytest.raises(ValueError):
            net_1d.fit(X, None)

    def test_net_2d_tensor_raises_error(self, net_2d):
        X = torch.arange(0, 8).view(4, 2).long()
        # We expect check_data to throw an exception
        # because we did not specify a custom data loader.
        with pytest.raises(ValueError):
            net_2d.fit(X, None)

    def test_net_1d_custom_loader(self, net_1d_custom_loader):
        X = torch.arange(0, 8).view(-1, 1).float()
        # throw away all callbacks since those may raise unrelated errors
        net_1d_custom_loader.initialize()
        net_1d_custom_loader.callbacks_ = []
        # Should not raise an exception.
        net_1d_custom_loader.partial_fit(X, None)

    def test_net_2d_custom_loader(self, net_2d_custom_loader):
        X = torch.arange(0, 8).view(4, 2).float()
        # throw away all callbacks since those may raise unrelated errors
        net_2d_custom_loader.initialize()
        net_2d_custom_loader.callbacks_ = []
        # Should not raise an exception.
        net_2d_custom_loader.partial_fit(X, None)


class TestNetWithDict:
    @pytest.fixture(scope='module')
    def module_cls(self):
        """Return a simple module that concatenates its 2 inputs in
        forward step.

        """
        class MyModule(nn.Module):
            def __init__(self):
                super(MyModule, self).__init__()
                self.dense = nn.Linear(20, 2)

            # pylint: disable=arguments-differ
            def forward(self, X0, X1):
                X = torch.cat((X0, X1), 1)
                X = F.softmax(self.dense(X), dim=-1)
                return X

        return MyModule

    @pytest.fixture(scope='module')
    def data(self):
        X, y = make_classification(1000, 20, n_informative=10, random_state=0)
        X = X.astype(np.float32)
        return X[:, :10], X[:, 10:], y

    @pytest.fixture(scope='module')
    def net_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.fixture(scope='module')
    def net(self, net_cls, module_cls):
        return net_cls(
            module_cls,
            max_epochs=2,
            lr=0.1,
        )

    def test_fit_predict_proba(self, net, data):
        X = {'X0': data[0], 'X1': data[1]}
        y = data[2]
        net.fit(X, y)
        y_proba = net.predict_proba(X)
        assert np.allclose(y_proba.sum(1), 1)

        # Issue #142: check that all batch sizes are consistent with
        # `net.batch_size`, even when the input type is a dictionary.
        # Note that we allow for different batch sizes as the total
        # number of samples may not be divisible by the batch size.
        batch_sizes = lambda n: set(sum(net.history[:, 'batches', :, n], []))
        train_batch_sizes = batch_sizes('train_batch_size')
        valid_batch_sizes = batch_sizes('valid_batch_size')
        assert net.batch_size in train_batch_sizes
        assert net.batch_size in valid_batch_sizes


class TestNetWithList:
    @pytest.fixture(scope='module')
    def module_cls(self):
        """Return a simple module that concatenates the input."""
        class MyModule(nn.Module):
            def __init__(self):
                super(MyModule, self).__init__()
                self.dense = nn.Linear(20, 2)

            # pylint: disable=arguments-differ
            def forward(self, X):
                X = torch.cat(X, 1)
                X = F.softmax(self.dense(X), dim=-1)
                return X

        return MyModule

    @pytest.fixture(scope='module')
    def data(self):
        X, y = make_classification(1000, 20, n_informative=10, random_state=0)
        X = X.astype(np.float32)
        return [X[:, :10], X[:, 10:]], y

    @pytest.fixture(scope='module')
    def net_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.fixture(scope='module')
    def net(self, net_cls, module_cls):
        return net_cls(
            module_cls,
            max_epochs=2,
            lr=0.1,
        )

    def test_fit_predict_proba(self, net, data):
        X, y = data
        net.fit(X, y)
        y_proba = net.predict_proba(X)
        assert np.allclose(y_proba.sum(1), 1)


@pytest.mark.skipif(not pandas_installed, reason='pandas is not installed')
class TestNetWithPandas:
    @pytest.fixture(scope='module')
    def module_cls(self):
        """Return a simple module that concatenates all input values
        in forward step.

        """
        class MyModule(nn.Module):
            def __init__(self):
                super(MyModule, self).__init__()
                self.dense = nn.Linear(20, 2)

            # pylint: disable=arguments-differ
            def forward(self, **X):
                X = torch.cat(list(X.values()), 1)
                X = F.softmax(self.dense(X), dim=-1)
                return X

        return MyModule

    @pytest.fixture(scope='module')
    def pd(self):
        import pandas as pd
        return pd

    @pytest.fixture(scope='module')
    def data(self, pd):
        X, y = make_classification(1000, 20, n_informative=10, random_state=0)
        X = X.astype(np.float32)
        df = pd.DataFrame(X, columns=map(str, range(X.shape[1])))
        return df, y

    @pytest.fixture(scope='module')
    def net_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.fixture(scope='module')
    def net(self, net_cls, module_cls):
        return net_cls(
            module_cls,
            max_epochs=2,
            lr=0.1,
        )

    def test_fit_predict_proba(self, net, data):
        X, y = data
        net.fit(X, y)
        y_proba = net.predict_proba(X)
        assert np.allclose(y_proba.sum(1), 1)


class TestNetWithTokenizers:
    """Huggingface tokenizers should work without special adjustments"""
    @pytest.fixture(scope='session')
    def tokenizer(self):
        transformers = pytest.importorskip('transformers')
        tokenizer = transformers.AutoTokenizer.from_pretrained('bert-base-uncased')
        return tokenizer

    @pytest.fixture(scope='session')
    def data(self, tokenizer):
        """A simple dataset that the model should be able to learn (or overfit)
        on

        """
        X = [paragraph for paragraph in unittest.__doc__.split('\n') if paragraph]
        Xt = tokenizer(
            X,
            max_length=12,
            padding='max_length',
            truncation=True,
            return_token_type_ids=False,
            return_tensors='pt',
        )
        y = np.array(['test' in x.lower() for x in X], dtype=np.int64)
        return Xt, y

    @pytest.fixture(scope='session')
    def module_cls(self, tokenizer):
        """Return a simple module using embedding + linear + softmax instead of
        a full-fledged BERT module.

        """
        class MyModule(nn.Module):
            def __init__(self):
                super().__init__()
                self.emb = nn.Embedding(tokenizer.vocab_size, 6)
                self.dense = nn.Linear(6, 2)
                self.sm = nn.Softmax(dim=-1)

            # pylint: disable=arguments-differ
            def forward(self, input_ids, attention_mask):
                assert input_ids.shape == attention_mask.shape
                X = self.emb(input_ids).mean(1)
                return self.sm(self.dense(X))

        return MyModule

    @pytest.fixture(scope='session')
    def net_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.fixture(scope='module')
    def net(self, net_cls, module_cls):
        return net_cls(
            module_cls,
            optimizer=torch.optim.Adam,
            max_epochs=5,
            batch_size=8,
            lr=0.1,
        )

    def test_fit_predict_proba(self, net, data):
        X, y = data
        net.fit(X, y)
        y_proba = net.predict_proba(X)
        assert np.allclose(y_proba.sum(1), 1)

        train_losses = net.history[:, 'train_loss']
        # make sure the network trained successfully with an arbitrary wide margin
        assert train_losses[0] > 5 * train_losses[-1]


class TestDataset:
    """Note: we don't need to test multi_indexing here, since that is
    already covered.

    """
    @pytest.fixture
    def dataset_cls(self):
        from skorch.dataset import Dataset
        return Dataset

    def test_len_correct(self, dataset_cls):
        pass

    def test_user_defined_len(self, dataset_cls):
        pass

    def test_inconsistent_lengths_raises(self, dataset_cls):
        pass

    def test_with_numpy_array(self, dataset_cls):
        pass

    def test_with_torch_tensor(self, dataset_cls):
        pass

    @pytest.mark.skipif(not pandas_installed, reason='pandas is not installed')
    def test_with_pandas_df(self, dataset_cls):
        pass

    @pytest.mark.skipif(not pandas_installed, reason='pandas is not installed')
    def test_with_pandas_series(self, dataset_cls):
        pass

    def test_with_dict(self, dataset_cls):
        pass

    def test_with_list_of_numpy_arrays(self, dataset_cls):
        pass

    @pytest.fixture
    def dataset_sparse_csr(self, dataset_cls):
        Xs = sparse.csr_matrix(np.zeros((10, 5)))
        return dataset_cls(Xs)

    @pytest.mark.parametrize('batch_size', [1, 3, 10, 17])
    def test_dataloader_with_sparse_csr(self, dataset_sparse_csr, batch_size):
        loader = DataLoader(dataset_sparse_csr, batch_size=batch_size)
        for Xb, _ in loader:
            assert is_torch_data_type(Xb)


class TestTrainSplitIsUsed:
    @pytest.fixture
    def iterator(self):
        """Return a simple iterator that yields the input data."""
        class Iterator:
            """An iterator that just yield the input data."""
            # pylint: disable=unused-argument
            def __init__(self, dataset, *args, **kwargs):
                self.dataset = dataset

            def __iter__(self):
                yield self.dataset.X, self.dataset.y

        return Iterator

    @pytest.fixture
    def data(self):
        X = torch.arange(0, 12, dtype=torch.float32).view(4, 3)
        y = torch.LongTensor([0, 1, 1, 0])
        return X, y

    @pytest.fixture
    def data_split(self, data):
        from skorch.dataset import Dataset

        X, y = data
        dataset_train = Dataset(X[:2], y[:2])
        dataset_valid = Dataset(X[2:], y[2:])
        return dataset_train, dataset_valid

    @pytest.fixture
    def module(self, classifier_module):
        return classifier_module

    @pytest.fixture
    def train_split(self, data_split):
        return Mock(side_effect=[data_split])

    @pytest.fixture
    def net_and_mock(self, module, data, train_split, iterator):
        """Return a NeuralNetClassifier with mocked train and
        validation step which save the args and kwargs the methods are
        calld with.

        """
        from skorch import NeuralNetClassifier

        X, y = data
        net = NeuralNetClassifier(
            module,
            module__input_units=3,
            max_epochs=1,
            iterator_train=iterator,
            iterator_valid=iterator,
            train_split=train_split
        )
        net.initialize()
        net.callbacks_ = []

        mock = Mock()

        def decorator(func):
            def wrapper(*args, **kwargs):
                mock(*args, **kwargs)
                func.__dict__['mock_'] = mock
                return func(*args[1:], **kwargs)
            return wrapper

        import types
        net.get_iterator = types.MethodType(decorator(net.get_iterator), net)
        return net.partial_fit(X, y), mock

    def test_steps_called_with_split_data(self, net_and_mock, data_split):
        mock = net_and_mock[1]
        assert mock.call_count == 2  # once for train, once for valid
        assert mock.call_args_list[0][0][1] == data_split[0]
        assert mock.call_args_list[0][1]['training'] is True
        assert mock.call_args_list[1][0][1] == data_split[1]
        assert mock.call_args_list[1][1]['training'] is False


class TestValidSplit:
    num_samples = 100

    @staticmethod
    def assert_datasets_equal(ds0, ds1):
        """Generic function to test equality of dataset values."""
        assert len(ds0) == len(ds1)
        # pylint: disable=consider-using-enumerate
        for i in range(len(ds0)):
            x0, y0 = ds0[i]
            x1, y1 = ds1[i]
            try:
                assert x0 == x1
            except (RuntimeError, ValueError):
                assert (x0 == x1).all()
            try:
                assert y0 == y1
            except (RuntimeError, ValueError):
                assert (y0 == y1).all()

    @pytest.fixture
    def dataset_cls(self):
        from skorch.dataset import Dataset
        return Dataset

    @pytest.fixture
    def data(self, dataset_cls):
        X = np.random.random((self.num_samples, 10))
        assert self.num_samples % 4 == 0
        y = np.repeat([0, 1, 2, 3], self.num_samples // 4)
        return dataset_cls(X, y)

    @pytest.fixture
    def valid_split_cls(self):
        from skorch.dataset import ValidSplit
        return ValidSplit

    def test_reproducible(self, valid_split_cls, data):
        dataset_train0, dataset_valid0 = valid_split_cls(5)(data)
        dataset_train1, dataset_valid1 = valid_split_cls(5)(data)
        self.assert_datasets_equal(dataset_train0, dataset_train1)
        self.assert_datasets_equal(dataset_valid0, dataset_valid1)

    @pytest.mark.parametrize('cv', [2, 4, 5, 10])
    def test_different_kfolds(self, valid_split_cls, cv, data):
        if self.num_samples % cv != 0:
            raise ValueError("Num samples not divisible by {}".format(cv))

        dataset_train, dataset_valid = valid_split_cls(cv)(data)
        assert len(dataset_train) + len(dataset_valid) == self.num_samples
        assert len(dataset_valid) == self.num_samples // cv

    @pytest.mark.parametrize('cv', [5, 0.2])
    def test_stratified(self, valid_split_cls, data, cv):
        num_expected = self.num_samples // 4
        y = np.hstack([np.repeat([0, 0, 0], num_expected),
                       np.repeat([1], num_expected)])
        data.y = y

        dataset_train, dataset_valid = valid_split_cls(
            cv, stratified=True)(data, y)
        y_train = data_from_dataset(dataset_train)[1]
        y_valid = data_from_dataset(dataset_valid)[1]

        assert y_train.sum() == 0.8 * num_expected
        assert y_valid.sum() == 0.2 * num_expected

    @pytest.mark.parametrize('cv', [0.1, 0.2, 0.5, 0.75])
    def test_different_fractions(self, valid_split_cls, cv, data):
        if not (self.num_samples * cv).is_integer() != 0:
            raise ValueError("Num samples cannot be evenly distributed for "
                             "fraction {}".format(cv))

        dataset_train, dataset_valid = valid_split_cls(cv)(data)
        assert len(dataset_train) + len(dataset_valid) == self.num_samples
        assert len(dataset_valid) == self.num_samples * cv

    @pytest.mark.parametrize('cv', [0.1, 0.2, 0.5, 0.75])
    def test_fraction_no_y(self, valid_split_cls, data, cv):
        if not (self.num_samples * cv).is_integer() != 0:
            raise ValueError("Num samples cannot be evenly distributed for "
                             "fraction {}".format(cv))

        m = int(cv * self.num_samples)
        n = int((1 - cv) * self.num_samples)
        dataset_train, dataset_valid = valid_split_cls(
            cv, stratified=False)(data, None)
        assert len(dataset_valid) == m
        assert len(dataset_train) == n

    def test_fraction_no_classifier(self, valid_split_cls, data):
        y = np.random.random(self.num_samples)
        data.y = y

        cv = 0.2
        m = int(cv * self.num_samples)
        n = int((1 - cv) * self.num_samples)
        dataset_train, dataset_valid = valid_split_cls(
            cv, stratified=False)(data, y)

        assert len(dataset_valid) == m
        assert len(dataset_train) == n

    @pytest.mark.parametrize('cv', [0, -0.001, -0.2, -3])
    def test_bad_values_raise(self, valid_split_cls, cv):
        with pytest.raises(ValueError) as exc:
            valid_split_cls(cv)

        expected = ("Numbers less than 0 are not allowed for cv "
                    "but ValidSplit got {}".format(cv))
        assert exc.value.args[0] == expected

    @pytest.mark.parametrize('cv', [5, 0.2])
    def test_not_stratified(self, valid_split_cls, data, cv):
        num_expected = self.num_samples // 4
        y = np.hstack([np.repeat([0, 0, 0], num_expected),
                       np.repeat([1], num_expected)])
        data.y = y

        dataset_train, dataset_valid = valid_split_cls(
            cv, stratified=False)(data, y)
        y_train = data_from_dataset(dataset_train)[1]
        y_valid = data_from_dataset(dataset_valid)[1]

        # when not stratified, we cannot know the distribution of targets
        assert y_train.sum() + y_valid.sum() == num_expected

    def test_predefined_split(self, valid_split_cls, data):
        from sklearn.model_selection import PredefinedSplit
        indices = (data.y > 0).astype(int)
        split = PredefinedSplit(indices)

        dataset_train, dataset_valid = valid_split_cls(split)(data)
        y_train = data_from_dataset(dataset_train)[1]
        y_valid = data_from_dataset(dataset_valid)[1]

        assert (y_train > 0).all()
        assert (y_valid == 0).all()

    def test_with_y_none(self, valid_split_cls, data):
        data.y = None
        m = self.num_samples // 5
        n = self.num_samples - m
        dataset_train, dataset_valid = valid_split_cls(5)(data)

        assert len(dataset_train) == n
        assert len(dataset_valid) == m

        y_train = data_from_dataset(dataset_train)[1]
        y_valid = data_from_dataset(dataset_valid)[1]

        assert y_train is None
        assert y_valid is None

    def test_with_torch_tensors(self, valid_split_cls, data):
        data.X = to_tensor(data.X, device='cpu')
        data.y = to_tensor(data.y, device='cpu')
        m = self.num_samples // 5
        n = self.num_samples - m
        dataset_train, dataset_valid = valid_split_cls(5)(data)

        assert len(dataset_valid) == m
        assert len(dataset_train) == n

    def test_with_torch_tensors_and_stratified(self, valid_split_cls, data):
        num_expected = self.num_samples // 4
        data.X = to_tensor(data.X, device='cpu')
        y = np.hstack([np.repeat([0, 0, 0], num_expected),
                       np.repeat([1], num_expected)])
        data.y = to_tensor(y, device='cpu')

        dataset_train, dataset_valid = valid_split_cls(5, stratified=True)(data, y)
        y_train = data_from_dataset(dataset_train)[1]
        y_valid = data_from_dataset(dataset_valid)[1]

        assert y_train.sum() == 0.8 * num_expected
        assert y_valid.sum() == 0.2 * num_expected

    def test_with_list_of_arrays(self, valid_split_cls, data):
        data.X = [data.X, data.X]
        m = self.num_samples // 5
        n = self.num_samples - m

        dataset_train, dataset_valid = valid_split_cls(5)(data)
        X_train, y_train = data_from_dataset(dataset_train)
        X_valid, y_valid = data_from_dataset(dataset_valid)

        assert len(X_train[0]) == len(X_train[1]) == len(y_train) == n
        assert len(X_valid[0]) == len(X_valid[1]) == len(y_valid) == m

    def test_with_dict(self, valid_split_cls, data):
        data.X = {'1': data.X, '2': data.X}
        dataset_train, dataset_valid = valid_split_cls(5)(data)

        m = self.num_samples // 5
        n = self.num_samples - m

        X_train, y_train = data_from_dataset(dataset_train)
        X_valid, y_valid = data_from_dataset(dataset_valid)

        assert len(X_train['1']) == len(X_train['2']) == len(y_train) == n
        assert len(X_valid['1']) == len(X_valid['2']) == len(y_valid) == m

    @pytest.mark.skipif(not pandas_installed, reason='pandas is not installed')
    def test_with_pandas(self, valid_split_cls, data):
        import pandas as pd

        data.X = pd.DataFrame(
            data.X,
            columns=[str(i) for i in range(data.X.shape[1])],
        )
        dataset_train, dataset_valid = valid_split_cls(5)(data)

        m = self.num_samples // 5
        X_train, y_train = data_from_dataset(dataset_train)
        X_valid, y_valid = data_from_dataset(dataset_valid)

        assert len(X_train) + len(X_valid) == self.num_samples
        assert len(y_train) + len(y_valid) == self.num_samples
        assert len(X_valid) == len(y_valid) == m

    def test_y_str_val_stratified(self, valid_split_cls, data):
        y = np.array(['a', 'a', 'a', 'b'] * (self.num_samples // 4))
        if len(data.X) != len(y):
            raise ValueError
        data.y = y

        dataset_train, dataset_valid = valid_split_cls(
            5, stratified=True)(data, y)
        y_train = data_from_dataset(dataset_train)[1]
        y_valid = data_from_dataset(dataset_valid)[1]

        assert np.isclose(np.mean(y_train == 'b'), 0.25)
        assert np.isclose(np.mean(y_valid == 'b'), 0.25)

    def test_y_list_of_arr_does_not_raise(self, valid_split_cls, data):
        y = [np.zeros(self.num_samples), np.ones(self.num_samples)]
        data.y = y
        valid_split_cls(5, stratified=False)(data)

    def test_y_list_of_arr_stratified(self, valid_split_cls, data):
        y = [np.zeros(self.num_samples), np.ones(self.num_samples)]
        data.y = y
        with pytest.raises(ValueError) as exc:
            valid_split_cls(5, stratified=True)(data, y)

        expected = "Stratified CV requires explicitly passing a suitable y."
        assert exc.value.args[0] == expected

    def test_y_dict_does_not_raise(self, valid_split_cls, data):
        y = {'a': np.zeros(self.num_samples), 'b': np.ones(self.num_samples)}
        data.y = y

        valid_split_cls(5, stratified=False)(data)

    def test_y_dict_stratified_raises(self, valid_split_cls, data):
        X = data[0]
        y = {'a': np.zeros(len(X)), 'b': np.ones(len(X))}

        with pytest.raises(ValueError):
            # an sklearn error is raised
            valid_split_cls(5, stratified=True)(X, y)

    @pytest.mark.parametrize('cv', [5, 0.2])
    @pytest.mark.parametrize('X', [np.zeros((100, 10)), torch.zeros((100, 10))])
    def test_y_none_stratified(self, valid_split_cls, data, cv, X):
        data.X = X
        with pytest.raises(ValueError) as exc:
            valid_split_cls(cv, stratified=True)(data, None)

        expected = "Stratified CV requires explicitly passing a suitable y."
        assert exc.value.args[0] == expected

    def test_shuffle_split_reproducible_with_random_state(
            self, valid_split_cls, dataset_cls):
        n = self.num_samples
        X, y = np.random.random((n, 10)), np.random.randint(0, 10, size=n)
        cv = valid_split_cls(0.2, stratified=False)

        dst0, dsv0 = cv(dataset_cls(X, y))
        dst1, dsv1 = cv(dataset_cls(X, y))

        Xt0, yt0 = data_from_dataset(dst0)
        Xv0, yv0 = data_from_dataset(dsv0)
        Xt1, yt1 = data_from_dataset(dst1)
        Xv1, yv1 = data_from_dataset(dsv1)

        assert not np.allclose(Xt0, Xt1)
        assert not np.allclose(Xv0, Xv1)
        assert not np.allclose(yt0, yt1)
        assert not np.allclose(yv0, yv1)

    def test_group_kfold(self, valid_split_cls, data):
        from sklearn.model_selection import GroupKFold

        X, y = data.X, data.y
        n = self.num_samples // 2
        groups = np.asarray(
            [0 for _ in range(n)] + [1 for _ in range(self.num_samples - n)])

        dataset_train, dataset_valid = valid_split_cls(
            GroupKFold(n_splits=2))(data, groups=groups)
        X_train, y_train = data_from_dataset(dataset_train)
        X_valid, y_valid = data_from_dataset(dataset_valid)

        assert np.allclose(X[:n], X_train)
        assert np.allclose(y[:n], y_train)
        assert np.allclose(X[n:], X_valid)
        assert np.allclose(y[n:], y_valid)

    def test_random_state_not_used_raises(self, valid_split_cls):
        # Since there is no randomness involved, raise a ValueError when
        # random_state is set, same as sklearn is now doing.
        msg = (
            r"Setting a random_state has no effect since cv is not a float. "
            r"You should leave random_state to its default \(None\), or set cv "
            r"to a float value."
        )
        with pytest.raises(ValueError, match=msg):
            valid_split_cls(5, random_state=0)

    def test_random_state_and_float_does_not_raise(self, valid_split_cls):
        valid_split_cls(0.5, random_state=0)  # does not raise

    def test_cvsplit_deprecation(self):
        from skorch.dataset import CVSplit
        with pytest.warns(
            DeprecationWarning,
            match="is deprecated, use the new name ValidSplit instead",
        ):
            CVSplit()



"""Contains learning rate scheduler callbacks"""

import sys

# pylint: disable=unused-import
import warnings

import numpy as np
import torch
from torch.optim.lr_scheduler import _LRScheduler
from torch.optim.lr_scheduler import CosineAnnealingLR
from torch.optim.lr_scheduler import ExponentialLR
from torch.optim.lr_scheduler import LambdaLR
from torch.optim.lr_scheduler import MultiStepLR
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.optim.lr_scheduler import StepLR

try:
    from torch.optim.lr_scheduler import CyclicLR as TorchCyclicLR
except ImportError:
    # Backward compatibility with torch >= 1.0 && < 1.1
    TorchCyclicLR = None
from torch.optim.optimizer import Optimizer
from skorch.callbacks import Callback

__all__ = ['LRScheduler', 'WarmRestartLR']


def _check_lr(name, optimizer, lr):
    """Return one learning rate for each param group."""
    n = len(optimizer.param_groups)
    if not isinstance(lr, (list, tuple)):
        return lr * np.ones(n)

    if len(lr) != n:
        raise ValueError("{} lr values were passed for {} but there are "
                         "{} param groups.".format(n, name, len(lr)))
    return np.array(lr)


class LRScheduler(Callback):
    """Callback that sets the learning rate of each
    parameter group according to some policy.

    Parameters
    ----------

    policy : str or _LRScheduler class (default='WarmRestartLR')
      Learning rate policy name or scheduler to be used.

    monitor : str or callable (default=None)
      Value of the history to monitor or function/callable. In
      the latter case, the callable receives the net instance as
      argument and is expected to return the score (float) used to
      determine the learning rate adjustment.

    event_name: str, (default='event_lr')
      Name of event to be placed in history when the scheduler takes a step.
      Pass ``None`` to disable placing events in history.
      **Note:** This feature works only for pytorch version >=1.4

    step_every: str, (default='epoch')
      Value for when to apply the learning scheduler step. Can be either 'batch'
       or 'epoch'.

    kwargs
      Additional arguments passed to the lr scheduler.

    """

    def __init__(self,
                 policy='WarmRestartLR',
                 monitor='train_loss',
                 event_name="event_lr",
                 step_every='epoch',
                 **kwargs):
        self.policy = policy
        self.monitor = monitor
        self.event_name = event_name
        self.step_every = step_every
        vars(self).update(kwargs)

    def simulate(self, steps, initial_lr):
        """
        Simulates the learning rate scheduler.

        Parameters
        ----------
        steps: int
          Number of steps to simulate

        initial_lr: float
          Initial learning rate

        Returns
        -------
        lrs: numpy ndarray
          Simulated learning rates

        """
        test = torch.ones(1, requires_grad=True)
        opt = torch.optim.SGD([{'params': test, 'lr': initial_lr}])
        policy_cls = self._get_policy_cls()
        sch = policy_cls(opt, **self.kwargs)

        lrs = []
        for _ in range(steps):
            opt.step()  # suppress warning about .step call order
            lrs.append(opt.param_groups[0]['lr'])
            sch.step()

        return np.array(lrs)

    def initialize(self):
        self.policy_ = self._get_policy_cls()
        self.lr_scheduler_ = None
        self.batch_idx_ = 0
        return self

    def _get_policy_cls(self):
        if isinstance(self.policy, str):
            return getattr(sys.modules[__name__], self.policy)
        return self.policy

    @property
    def kwargs(self):
        # These are the parameters that are passed to the
        # scheduler. Parameters that don't belong there must be
        # excluded.
        excluded = ('policy', 'monitor', 'event_name', 'step_every')
        kwargs = {key: val for key, val in vars(self).items()
                  if not (key in excluded or key.endswith('_'))}
        return kwargs

    def on_train_begin(self, net, **kwargs):
        if net.history:
            try:
                self.batch_idx_ = sum(net.history[:, 'train_batch_count'])
            except KeyError:
                self.batch_idx_ = sum(len(b) for b in net.history[:, 'batches'])
        self.lr_scheduler_ = self._get_scheduler(
            net, self.policy_, **self.kwargs
        )

    def on_epoch_end(self, net, **kwargs):
        if self.step_every != 'epoch':
            return
        if isinstance(self.lr_scheduler_, ReduceLROnPlateau):
            if callable(self.monitor):
                score = self.monitor(net)
            else:
                try:
                    score = net.history[-1, self.monitor]
                except KeyError as e:
                    raise ValueError(
                        f"'{self.monitor}' was not found in history. A "
                        f"Scoring callback with name='{self.monitor}' "
                        "should be placed before the LRScheduler callback"
                    ) from e

            self.lr_scheduler_.step(score)
            # ReduceLROnPlateau does not expose the current lr so it can't be recorded
        else:
            if self.event_name is not None and hasattr(
                    self.lr_scheduler_, "get_last_lr"):
                net.history.record(self.event_name,
                                   self.lr_scheduler_.get_last_lr()[0])
            self.lr_scheduler_.step()

    def on_batch_end(self, net, training, **kwargs):
        if not training or self.step_every != 'batch':
            return
        if self.event_name is not None and hasattr(
                self.lr_scheduler_, "get_last_lr"):
            net.history.record_batch(self.event_name,
                                     self.lr_scheduler_.get_last_lr()[0])
        self.lr_scheduler_.step()
        self.batch_idx_ += 1

    def _get_scheduler(self, net, policy, **scheduler_kwargs):
        """Return scheduler, based on indicated policy, with appropriate
        parameters.
        """
        if policy not in [ReduceLROnPlateau] and \
                'last_epoch' not in scheduler_kwargs:
            last_epoch = len(net.history) - 1
            scheduler_kwargs['last_epoch'] = last_epoch

        return policy(net.optimizer_, **scheduler_kwargs)


class WarmRestartLR(_LRScheduler):
    """Stochastic Gradient Descent with Warm Restarts (SGDR) scheduler.

    This scheduler sets the learning rate of each parameter group
    according to stochastic gradient descent with warm restarts (SGDR)
    policy. This policy simulates periodic warm restarts of SGD, where
    in each restart the learning rate is initialize to some value and is
    scheduled to decrease.

    Parameters
    ----------
    optimizer : torch.optimizer.Optimizer instance.
      Optimizer algorithm.

    min_lr : float or list of float (default=1e-6)
      Minimum allowed learning rate during each period for all
      param groups (float) or each group (list).

    max_lr : float or list of float (default=0.05)
      Maximum allowed learning rate during each period for all
      param groups (float) or each group (list).

    base_period : int (default=10)
      Initial restart period to be multiplied at each restart.

    period_mult : int (default=2)
      Multiplicative factor to increase the period between restarts.

    last_epoch : int (default=-1)
      The index of the last valid epoch.

    References
    ----------
    .. [1] Ilya Loshchilov and Frank Hutter, 2017, "Stochastic Gradient
        Descent with Warm Restarts,". "ICLR"
        `<https://arxiv.org/pdf/1608.03983.pdf>`_

    """

    def __init__(
            self, optimizer,
            min_lr=1e-6,
            max_lr=0.05,
            base_period=10,
            period_mult=2,
            last_epoch=-1
    ):
        self.min_lr = _check_lr('min_lr', optimizer, min_lr)
        self.max_lr = _check_lr('max_lr', optimizer, max_lr)
        self.base_period = base_period
        self.period_mult = period_mult
        super(WarmRestartLR, self).__init__(optimizer, last_epoch)

    def _get_current_lr(self, min_lr, max_lr, period, epoch):
        return min_lr + 0.5 * (max_lr - min_lr) * (
            1 + np.cos(epoch * np.pi / period))

    def get_lr(self):
        epoch_idx = float(self.last_epoch)
        current_period = float(self.base_period)
        while epoch_idx / current_period > 1.0:
            epoch_idx -= current_period + 1
            current_period *= self.period_mult

        current_lrs = self._get_current_lr(
            self.min_lr,
            self.max_lr,
            current_period,
            epoch_idx
        )
        return current_lrs.tolist()



"""Tests for classifier.py

Only contains tests that are specific for classifier subclasses.

"""

from unittest.mock import Mock

import numpy as np
import pytest
import torch
from scipy.special import expit
from sklearn.base import clone
from torch import nn

from skorch.tests.conftest import INFERENCE_METHODS


class TestNeuralNet:
    @pytest.fixture(scope='module')
    def data(self, classifier_data):
        return classifier_data

    @pytest.fixture(scope='module')
    def dummy_callback(self):
        from skorch.callbacks import Callback
        cb = Mock(spec=Callback)
        # make dummy behave like an estimator
        cb.get_params.return_value = {}
        cb.set_params = lambda **kwargs: cb
        return cb

    @pytest.fixture(scope='module')
    def net_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.fixture(scope='module')
    def module_cls(self, classifier_module):
        return classifier_module

    @pytest.fixture(scope='module')
    def net(self, net_cls, module_cls, dummy_callback):
        return net_cls(
            module_cls,
            callbacks=[('dummy', dummy_callback)],
            max_epochs=10,
            lr=0.1,
        )

    @pytest.fixture(scope='module')
    def net_fit(self, net, data):
        # Careful, don't call additional fits on this, since that would have
        # side effects on other tests.
        X, y = data
        return net.fit(X, y)

    def test_clone(self, net_fit):
        clone(net_fit)

    def test_predict_and_predict_proba(self, net_fit, data):
        X = data[0]

        y_proba = net_fit.predict_proba(X)
        assert np.allclose(y_proba.sum(1), 1, rtol=1e-5)

        y_pred = net_fit.predict(X)
        assert np.allclose(np.argmax(y_proba, 1), y_pred, rtol=1e-5)

    def test_score(self, net_fit, data):
        X, y = data
        accuracy = net_fit.score(X, y)
        assert 0. <= accuracy <= 1.

    # classifier-specific test
    def test_takes_log_with_nllloss(self, net_cls, module_cls, data):
        net = net_cls(module_cls, criterion=nn.NLLLoss, max_epochs=1)
        net.initialize()

        mock_loss = Mock(side_effect=nn.NLLLoss())
        net.criterion_.forward = mock_loss
        net.partial_fit(*data)  # call partial_fit to avoid re-initialization

        # check that loss was called with log-probabilities
        for (y_log, _), _ in mock_loss.call_args_list:
            assert (y_log < 0).all()
            y_proba = torch.exp(y_log)
            assert torch.isclose(torch.ones(len(y_proba)), y_proba.sum(1)).all()

    # classifier-specific test
    def test_takes_no_log_without_nllloss(self, net_cls, module_cls, data):
        net = net_cls(module_cls, criterion=nn.CrossEntropyLoss, max_epochs=1)
        net.initialize()

        mock_loss = Mock(side_effect=nn.NLLLoss())
        net.criterion_.forward = mock_loss
        net.partial_fit(*data)  # call partial_fit to avoid re-initialization

        # check that loss was called with raw probabilities
        for (y_out, _), _ in mock_loss.call_args_list:
            assert not (y_out < 0).all()
            assert torch.isclose(torch.ones(len(y_out)), y_out.sum(1)).all()

    # classifier-specific test
    def test_high_learning_rate(self, net_cls, module_cls, data):
        # regression test for nan loss with high learning rates issue #481
        net = net_cls(module_cls, max_epochs=2, lr=2, optimizer=torch.optim.Adam)
        net.fit(*data)
        assert np.any(~np.isnan(net.history[:, 'train_loss']))

    def test_binary_classes_set_by_default(self, net_cls, module_cls, data):
        net = net_cls(module_cls).fit(*data)
        assert (net.classes_ == [0, 1]).all()

    def test_non_binary_classes_set_by_default(self, net_cls, module_cls, data):
        X = data[0]
        y = np.arange(len(X)) % 10
        net = net_cls(module_cls, max_epochs=0).fit(X, y)
        assert (net.classes_ == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).all()

    def test_classes_data_torch_tensor(self, net_cls, module_cls, data):
        X = torch.as_tensor(data[0])
        y = torch.as_tensor(np.arange(len(X)) % 10)

        net = net_cls(module_cls, max_epochs=0).fit(X, y)
        assert (net.classes_ == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).all()

    def test_classes_with_gaps(self, net_cls, module_cls, data):
        X = data[0]
        y = np.arange(len(X)) % 10
        y[(y == 0) | (y == 5)] = 4  # remove classes 0 and 5
        net = net_cls(module_cls, max_epochs=0).fit(X, y)
        assert (net.classes_ == [1, 2, 3, 4, 6, 7, 8, 9]).all()

    def test_pass_classes_explicitly_overrides(self, net_cls, module_cls, data):
        net = net_cls(module_cls, max_epochs=0, classes=['foo', 'bar']).fit(*data)
        assert net.classes_ == ['foo', 'bar']

    @pytest.mark.parametrize('classes', [[], np.array([])])
    def test_pass_empty_classes_raises(
            self, net_cls, module_cls, data, classes):
        net = net_cls(
            module_cls, max_epochs=0, classes=classes).fit(*data).fit(*data)
        with pytest.raises(AttributeError) as exc:
            net.classes_  # pylint: disable=pointless-statement

        msg = exc.value.args[0]
        expected = "NeuralNetClassifier has no attribute 'classes_'"
        assert msg == expected

    def test_with_calibrated_classifier_cv(self, net_fit, data):
        from sklearn.calibration import CalibratedClassifierCV
        cccv = CalibratedClassifierCV(net_fit, cv=2)
        cccv.fit(*data)


class TestNeuralNetBinaryClassifier:
    @pytest.fixture(scope='module')
    def data(self, classifier_data):
        X, y = classifier_data
        return X, y.astype('float32')

    @pytest.fixture(scope='module')
    def module_cls(self):
        from skorch.toy import make_binary_classifier
        return make_binary_classifier(
            input_units=20,
            hidden_units=10,
            output_units=1,
            num_hidden=1,
            dropout=0,
        )

    @pytest.fixture(scope='module')
    def net_cls(self):
        from skorch.classifier import NeuralNetBinaryClassifier
        return NeuralNetBinaryClassifier

    @pytest.fixture(scope='module')
    def net(self, net_cls, module_cls):
        return net_cls(
            module_cls,
            max_epochs=1,
            lr=1,
        )

    @pytest.fixture(scope='module')
    def net_fit(self, net, data):
        # Careful, don't call additional fits on this, since that would have
        # side effects on other tests.
        net.set_params(max_epochs=10)
        X, y = data
        net.fit(X, y)
        net.set_params(max_epochs=1)
        return net

    def test_fit(self, net_fit):
        # fitting does not raise anything
        pass

    def test_clone(self, net_fit):
        clone(net_fit)

    @pytest.mark.parametrize('method', INFERENCE_METHODS)
    def test_not_fitted_raises(self, net_cls, module_cls, data, method):
        from skorch.exceptions import NotInitializedError
        net = net_cls(module_cls)
        X = data[0]
        with pytest.raises(NotInitializedError) as exc:
            # we call `list` because `forward_iter` is lazy
            list(getattr(net, method)(X))

        msg = ("This NeuralNetBinaryClassifier instance is not initialized "
               "yet. Call 'initialize' or 'fit' with appropriate arguments "
               "before using this method.")
        assert exc.value.args[0] == msg

    def test_net_learns(self, net_cls, module_cls, data):
        X, y = data
        net = net_cls(
            module_cls,
            max_epochs=10,
            lr=1,
            batch_size=64,
        )
        net.fit(X, y)

        train_losses = net.history[:, 'train_loss']
        assert train_losses[0] > 1.3 * train_losses[-1]

        valid_acc = net.history[-1, 'valid_acc']
        assert valid_acc > 0.65

    def test_batch_size_one(self, net_cls, module_cls, data):
        X, y = data
        net = net_cls(
            module_cls,
            max_epochs=1,
            batch_size=1,
        )
        net.fit(X, y)

    def test_history_default_keys(self, net_fit):
        expected_keys = {
            'train_loss', 'valid_loss', 'epoch', 'dur', 'batches', 'valid_acc'
        }
        for row in net_fit.history:
            assert expected_keys.issubset(row)

    @pytest.mark.parametrize('threshold', [0, 0.25, 0.5, 0.75, 1])
    def test_predict_predict_proba(self, net, data, threshold):
        X, y = data
        net.threshold = threshold
        net.fit(X, y)

        y_pred_proba = net.predict_proba(X)
        assert y_pred_proba.shape == (X.shape[0], 2)

        # The tests below check that we don't accidentally apply sigmoid twice,
        # which would result in probabilities constrained to [expit(-1),
        # expit(1)]. The lower bound is not expit(0), as one may think at first,
        # because we create the probabilities as:
        # torch.stack((1 - prob, prob), 1)
        # So the lowest value that could be achieved by applying sigmoid twice
        # is 1 - expit(1), which is equal to expit(-1).
        prob_min, prob_max = expit(-1), expit(1)
        assert (y_pred_proba < prob_min).any()
        assert (y_pred_proba > prob_max).any()

        y_pred_exp = (y_pred_proba[:, 1] > threshold).astype('uint8')

        y_pred_actual = net.predict(X)
        assert np.allclose(y_pred_exp, y_pred_actual)

    def test_score(self, net_fit, data):
        X, y = data
        accuracy = net_fit.score(X, y)
        assert 0. <= accuracy <= 1.

    def test_fit_with_dataset_and_y_none(self, net_cls, module_cls, data):
        from skorch.dataset import Dataset

        # deactivate train split since it requires y
        net = net_cls(module_cls, train_split=False, max_epochs=1)
        X, y = data
        dataset = Dataset(X, y)
        assert net.fit(dataset, y=None)

    def test_target_2d_raises(self, net, data):
        X, y = data
        with pytest.raises(ValueError) as exc:
            net.fit(X, y[:, None])

        assert exc.value.args[0] == (
            "The target data should be 1-dimensional.")

    def test_custom_loss_does_not_call_sigmoid(
            self, net_cls, data, module_cls, monkeypatch):
        mock = Mock(side_effect=lambda x: x)
        monkeypatch.setattr(torch, "sigmoid", mock)

        # add a custom nonlinearity - note that the output must return
        # a 2d array from a 1d vector to conform to the required
        # y_proba
        def nonlin(x):
            return torch.stack((1 - x, x), 1)

        net = net_cls(module_cls, max_epochs=1, lr=0.1, criterion=nn.MSELoss,
                      predict_nonlinearity=nonlin)
        X, y = data
        net.fit(X, y)

        net.predict_proba(X)
        assert mock.call_count == 0

    def test_default_loss_does_call_sigmoid(
            self, net_cls, data, module_cls, monkeypatch):
        mock = Mock(side_effect=lambda x: x)
        monkeypatch.setattr(torch, "sigmoid", mock)

        net = net_cls(module_cls, max_epochs=1, lr=0.1)
        X, y = data
        net.fit(X, y)

        net.predict_proba(X)
        assert mock.call_count > 0

    def test_with_calibrated_classifier_cv(self, net_fit, data):
        from sklearn.calibration import CalibratedClassifierCV
        cccv = CalibratedClassifierCV(net_fit, cv=2)
        cccv.fit(*data)

    def test_grid_search_with_roc_auc(self, net_fit, data):
        from sklearn.model_selection import GridSearchCV
        search = GridSearchCV(
            net_fit,
            {'max_epochs': [1, 2]},
            refit=False,
            cv=3,
            scoring='roc_auc',
        )
        search.fit(*data)

    def test_module_output_not_1d(self, net_cls, data):
        from skorch.toy import make_classifier
        module = make_classifier(
            input_units=20,
            output_units=1,
        )  # the output will not be squeezed
        net = net_cls(module, max_epochs=1)
        net.fit(*data)  # does not raise

    def test_module_output_2d_raises(self, net_cls, data):
        from skorch.toy import make_classifier
        module = make_classifier(
            input_units=20,
            output_units=2,
        )
        net = net_cls(module, max_epochs=1)
        with pytest.raises(ValueError) as exc:
            net.fit(*data)

        msg = exc.value.args[0]
        expected = ("Expected module output to have shape (n,) or "
                    "(n, 1), got (128, 2) instead")
        assert msg == expected

    @pytest.fixture(scope='module')
    def net_with_bceloss(self, net_cls, module_cls, data):
        # binary classification should also work with BCELoss
        net = net_cls(
            module_cls,
            module__output_nonlin=torch.nn.Sigmoid(),
            criterion=torch.nn.BCELoss,
            lr=1,
        )
        X, y = data
        net.fit(X, y)
        return net

    def test_net_with_bceloss_learns(self, net_with_bceloss):
        train_losses = net_with_bceloss.history[:, 'train_loss']
        assert train_losses[0] > 1.3 * train_losses[-1]

    def test_predict_proba_with_bceloss(self, net_with_bceloss, data):
        X, _ = data
        y_proba = net_with_bceloss.predict_proba(X)

        assert y_proba.shape == (X.shape[0], 2)
        assert (y_proba >= 0).all()
        assert (y_proba <= 1).all()

        # The tests below check that we don't accidentally apply sigmoid twice,
        # which would result in probabilities constrained to [expit(-1),
        # expit(1)]. The lower bound is not expit(0), as one may think at first,
        # because we create the probabilities as:
        # torch.stack((1 - prob, prob), 1)
        # So the lowest value that could be achieved by applying sigmoid twice
        # is 1 - expit(1), which is equal to expit(-1).
        prob_min, prob_max = expit(-1), expit(1)
        assert (y_proba < prob_min).any()
        assert (y_proba > prob_max).any()

    def test_predict_with_bceloss(self, net_with_bceloss, data):
        X, _ = data

        y_pred_proba = net_with_bceloss.predict_proba(X)
        y_pred_exp = (y_pred_proba[:, 1] > net_with_bceloss.threshold).astype('uint8')
        y_pred_actual = net_with_bceloss.predict(X)
        assert np.allclose(y_pred_exp, y_pred_actual)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# skorch documentation build configuration file, created by
# sphinx-quickstart on Fri Oct 13 11:29:12 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath('..'))


#  on_rtd is whether we are on readthedocs.org, this line of code grabbed
#  from docs.readthedocs.org
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'numpydoc',
    'sphinx.ext.linkcode',
    'sphinx.ext.autosummary',
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.intersphinx',
]

numpydoc_class_members_toctree = False

intersphinx_mapping = {
    'pytorch': ('https://pytorch.org/docs/stable/', None),
    'sklearn': ('https://scikit-learn.org/stable/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'python': ('https://docs.python.org/3', None),
    'mlflow': ('https://mlflow.org/docs/latest/', None),
    'gpytorch': ('https://docs.gpytorch.ai/en/stable/', None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'skorch'
copyright = '2017, Marian Tietz, Daniel Nouri, Benjamin Bossan'
author = 'Marian Tietz, Daniel Nouri, Benjamin Bossan'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
# version = '0.1'
# The full version, including alpha/beta/rc tags.
# release = '0.1'
with open('../VERSION', 'r') as f:
    release = f.read().strip()
    version = release.rsplit('.', 1)[0]

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**tests**']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'default'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

def setup(app):
    app.add_css_file('css/my_theme.css')

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'skorchdoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'skorch.tex', 'skorch Documentation',
     'Marian Tietz, Daniel Nouri, Benjamin Bossan', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'skorch', 'skorch Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'skorch', 'skorch Documentation',
     author, 'skorch', 'One line description of project.',
     'Miscellaneous'),
]

# -- GitHub source code link ----------------------------------------------

# Functionality to build github source URI, taken from sklearn.

from operator import attrgetter
import inspect
import subprocess
from functools import partial

REVISION_CMD = 'git rev-parse --short HEAD'

def _get_git_revision():
    try:
        revision = subprocess.check_output(REVISION_CMD.split()).strip()
    except (subprocess.CalledProcessError, OSError):
        print('Failed to execute git to get revision')
        return None
    return revision.decode('utf-8')


def _linkcode_resolve(domain, info, package, url_fmt, revision):
    """Determine a link to online source for a class/method/function

    This is called by sphinx.ext.linkcode

    An example with a long-untouched module that everyone has
    >>> _linkcode_resolve('py', {'module': 'tty',
    ...                          'fullname': 'setraw'},
    ...                   package='tty',
    ...                   url_fmt='http://hg.python.org/cpython/file/'
    ...                           '{revision}/Lib/{package}/{path}#L{lineno}',
    ...                   revision='xxxx')
    'http://hg.python.org/cpython/file/xxxx/Lib/tty/tty.py#L18'
    """

    if revision is None:
        return
    if domain not in ('py', 'pyx'):
        return
    if not info.get('module') or not info.get('fullname'):
        return

    class_name = info['fullname'].split('.')[0]
    if type(class_name) != str:
        # Python 2 only
        class_name = class_name.encode('utf-8')
    module = __import__(info['module'], fromlist=[class_name])
    obj = attrgetter(info['fullname'])(module)

    try:
        fn = inspect.getsourcefile(obj)
    except Exception:
        fn = None
    if not fn:
        try:
            fn = inspect.getsourcefile(sys.modules[obj.__module__])
        except Exception:
            fn = None
    if not fn:
        return

    fn = os.path.relpath(fn,
                         start=os.path.dirname(__import__(package).__file__))
    try:
        lineno = inspect.getsourcelines(obj)[1]
    except Exception:
        lineno = ''
    return url_fmt.format(revision=revision, package=package,
                          path=fn, lineno=lineno)

def project_linkcode_resolve(domain, info):
    global _linkcode_git_revision
    return _linkcode_resolve(domain, info,
            package='skorch',
            revision=_linkcode_git_revision,
            url_fmt='https://github.com/skorch-dev/skorch/'
                    'blob/{revision}/'
                    '{package}/{path}#L{lineno}')

_linkcode_git_revision = _get_git_revision()

# The following is used by sphinx.ext.linkcode to provide links to github
linkcode_resolve = project_linkcode_resolve



"""Simple training script for a MLP classifier.

See accompanying README.md for more details.

"""

import pickle

import fire
import numpy as np
from sklearn.datasets import make_classification
from sklearn.feature_selection import SelectKBest
from sklearn.pipeline import FeatureUnion
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import Normalizer
from skorch import NeuralNetClassifier
import torch
from torch import nn

from skorch.helper import parse_args


np.random.seed(0)
torch.manual_seed(0)
torch.cuda.manual_seed(0)


# number of input features
N_FEATURES = 20

# number of classes
N_CLASSES = 2

# custom defaults for net
DEFAULTS_NET = {
    'batch_size': 256,
    'module__hidden_units': 30,
}

# custom defaults for pipeline
DEFAULTS_PIPE = {
    'scale__minmax__feature_range': (-1, 1),
    'net__batch_size': 256,
    'net__module__hidden_units': 30,
}


class MLPClassifier(nn.Module):
    """A simple multi-layer perceptron module.

    This can be adapted for usage in different contexts, e.g. binary
    and multi-class classification, regression, etc.

    Note: This docstring is used to create the help for the CLI.

    Parameters
    ----------
    hidden_units : int (default=10)
      Number of units in hidden layers.

    num_hidden : int (default=1)
      Number of hidden layers.

    nonlin : torch.nn.Module instance (default=torch.nn.ReLU())
      Non-linearity to apply after hidden layers.

    dropout : float (default=0)
      Dropout rate. Dropout is applied between layers.

    """
    def __init__(
            self,
            hidden_units=10,
            num_hidden=1,
            nonlin=nn.ReLU(),
            dropout=0,
    ):
        super().__init__()
        self.hidden_units = hidden_units
        self.num_hidden = num_hidden
        self.nonlin = nonlin
        self.dropout = dropout

        self.reset_params()

    def reset_params(self):
        """(Re)set all parameters."""
        units = [N_FEATURES]
        units += [self.hidden_units] * self.num_hidden
        units += [N_CLASSES]

        sequence = []
        for u0, u1 in zip(units, units[1:]):
            sequence.append(nn.Linear(u0, u1))
            sequence.append(self.nonlin)
            sequence.append(nn.Dropout(self.dropout))

        sequence = sequence[:-2]
        self.sequential = nn.Sequential(*sequence)

    def forward(self, X):
        return nn.Softmax(dim=-1)(self.sequential(X))


def get_data(n_samples=100):
    """Get synthetic classification data with n_samples samples."""
    X, y = make_classification(
        n_samples=n_samples,
        n_features=N_FEATURES,
        n_classes=N_CLASSES,
        random_state=0,
    )
    X = X.astype(np.float32)
    return X, y


def get_model(with_pipeline=False):
    """Get a multi-layer perceptron model.

    Optionally, put it in a pipeline that scales the data.

    """
    model = NeuralNetClassifier(MLPClassifier)
    if with_pipeline:
        model = Pipeline([
            ('scale', FeatureUnion([
                ('minmax', MinMaxScaler()),
                ('normalize', Normalizer()),
            ])),
            ('select', SelectKBest(k=N_FEATURES)),  # keep input size constant
            ('net', model),
        ])
    return model


def save_model(model, output_file):
    """Save model to output_file, if given"""
    if not output_file:
        return

    with open(output_file, 'wb') as f:
        pickle.dump(model, f)
    print("Saved model to file '{}'.".format(output_file))


def net(n_samples=100, output_file=None, **kwargs):
    """Train an MLP classifier on synthetic data.

    n_samples : int (default=100)
      Number of training samples

    output_file : str (default=None)
      If not None, file name used to save the model.

    kwargs : dict
      Additional model parameters.

    """

    model = get_model(with_pipeline=False)
    # important: wrap the model with the parsed arguments
    parsed = parse_args(kwargs, defaults=DEFAULTS_NET)
    model = parsed(model)

    X, y = get_data(n_samples=n_samples)
    print("Training MLP classifier")
    model.fit(X, y)

    save_model(model, output_file)


def pipeline(n_samples=100, output_file=None, **kwargs):
    """Train an MLP classifier in a pipeline on synthetic data.

    The pipeline scales the input data before passing it to the net.

    Note: This docstring is used to create the help for the CLI.

    Parameters
    ----------
    n_samples : int (default=100)
      Number of training samples

    output_file : str (default=None)
      If not None, file name used to save the model.

    kwargs : dict
      Additional model parameters.

    """

    model = get_model(with_pipeline=True)
    # important: wrap the model with the parsed arguments
    parsed = parse_args(kwargs, defaults=DEFAULTS_PIPE)
    model = parsed(model)

    X, y = get_data(n_samples=n_samples)
    print("Training MLP classifier in a pipeline")
    model.fit(X, y)

    save_model(model, output_file)


if __name__ == '__main__':
    # register 2 functions, "net" and "pipeline"
    fire.Fire({'net': net, 'pipeline': pipeline})



import itertools

import pytest


class TestAllCallbacks:
    @pytest.fixture
    def callbacks(self):
        """Return all callbacks"""
        import skorch.callbacks

        callbacks = []
        for name in dir(skorch.callbacks):
            attr = getattr(skorch.callbacks, name)
            # pylint: disable=unidiomatic-typecheck
            if not type(attr) is type:
                continue
            if issubclass(attr, skorch.callbacks.Callback):
                callbacks.append(attr)
        return callbacks

    @pytest.fixture
    def base_cls(self):
        from skorch.callbacks import Callback
        return Callback

    @pytest.fixture
    def on_x_methods(self):
        return [
            'on_train_begin',
            'on_train_end',
            'on_epoch_begin',
            'on_epoch_end',
            'on_batch_begin',
            'on_batch_end',
            'on_grad_computed',
        ]

    def test_on_x_methods_have_kwargs(self, callbacks, on_x_methods):
        import inspect
        for callback, method_name in itertools.product(
                callbacks, on_x_methods):
            method = getattr(callback, method_name)
            assert "kwargs" in inspect.signature(method).parameters

    def test_set_params_with_unknown_key_raises(self, base_cls):
        with pytest.raises(ValueError) as exc:
            base_cls().set_params(foo=123)
        msg = exc.value.args[0]

        # message contains "'" around variable name starting from sklearn 1.1
        assert (
            msg.startswith(
                "Invalid parameter foo for estimator <skorch.callbacks.base.Callback"
            )
            or msg.startswith(
                "Invalid parameter 'foo' for estimator <skorch.callbacks.base.Callback"
            )
        )



"""Tests for probabilistic.py"""

import copy
import pickle
import re

import numpy as np
import pytest
from sklearn.base import clone
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import Pipeline
import torch
from torch.testing import assert_allclose

from skorch._version import Version
from skorch.utils import is_torch_data_type
from skorch.utils import to_numpy


try:
    gpytorch = pytest.importorskip('gpytorch')
except AttributeError:
    pytest.skip("Incompatible gpytorch + torch version", allow_module_level=True)

# check that torch version is sufficiently high for gpytorch, otherwise skip
version_gpytorch = Version(gpytorch.__version__)
version_torch = Version(torch.__version__)
# TODO: remove if newer GPyTorch versions are released that no longer require
# the check.
if (version_gpytorch >= Version('1.9')) and (version_torch < Version('1.11')):
    pytest.skip("Incompatible gpytorch + torch version", allow_module_level=True)
elif (version_gpytorch >= Version('1.7')) and (version_torch < Version('1.10')):
    pytest.skip("Incompatible gpytorch + torch version", allow_module_level=True)


def get_batch_size(dist):
    """Return the shape of the distribution

    The method/attribute required to determine the shape depends on the kind of
    distrubtion.

    """
    shape = getattr(dist, 'shape', None)
    if shape:
        return shape[0]

    get_base_samples = getattr(dist, 'get_base_samples', None)
    if get_base_samples:
        return get_base_samples().shape[0]

    shape = getattr(dist, 'batch_shape', None)
    if shape:
        return shape[0]

    raise AttributeError(f"Could not determine shape of {dist}")


# PyTorch Modules are defined on the module root to make them pickleable.

class RbfModule(gpytorch.models.ExactGP):
    """Simple exact GP regression module"""
    def __init__(self, likelihood):
        super().__init__(None, None, likelihood)
        self.mean_module = gpytorch.means.ConstantMean()
        self.covar_module = gpytorch.kernels.RBFKernel()

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


class VariationalRegressionModule(gpytorch.models.ApproximateGP):
    """GP regression for variational inference"""
    def __init__(self, inducing_points, eps=1e-6):
        variational_distribution = gpytorch.variational.CholeskyVariationalDistribution(
            inducing_points.size(0))
        variational_strategy = gpytorch.variational.VariationalStrategy(
            self,
            inducing_points,
            variational_distribution,
            learn_inducing_locations=True,
        )
        super().__init__(variational_strategy)
        self.mean_module = gpytorch.means.ConstantMean()
        self.covar_module = gpytorch.kernels.ScaleKernel(
            gpytorch.kernels.RBFKernel(eps=eps))

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


class VariationalBinaryClassificationModule(gpytorch.models.ApproximateGP):
    """GP classification for variational inference"""
    def __init__(self, inducing_points, eps=1e-6):
        variational_distribution = gpytorch.variational.CholeskyVariationalDistribution(
            inducing_points.size(0))
        variational_strategy = gpytorch.variational.UnwhitenedVariationalStrategy(
            self,
            inducing_points,
            variational_distribution,
            learn_inducing_locations=False,
        )
        super().__init__(variational_strategy)
        self.mean_module = gpytorch.means.ConstantMean()
        self.covar_module = gpytorch.kernels.ScaleKernel(
            gpytorch.kernels.RBFKernel(eps=eps))

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        latent_pred = gpytorch.distributions.MultivariateNormal(mean_x, covar_x)
        return latent_pred


class BaseProbabilisticTests:
    """Base class for all GP estimators.

    This class defined all fixtures, most of which need to be implemented by the
    respective subclass, as well as all the tests. The tests take care of using
    attributes and properties that are true for all sorts of GPs (e.g. only
    using parameters shared by all likelihoods).

    """
    #####################
    # testing functions #
    #####################

    @staticmethod
    def assert_values_differ(x):
        x = to_numpy(x)
        assert len(np.unique(x)) > 1

    ##########################
    # constants and fixtures #
    ##########################

    @property
    def n_samples(self):
        # expects int
        return NotImplementedError

    @property
    def n_targets(self):
        # expects int
        raise NotImplementedError

    @property
    def supports_predict_proba(self):
        # expects bool
        raise NotImplementedError

    @property
    def supports_return_std(self):
        # expects bool
        # This only checks if the argument is allowed by predict, not whether it
        # actually implements a solution
        raise NotImplementedError

    @property
    def supports_return_cov(self):
        # expects bool
        # This only checks if the argument is allowed by predict, not whether it
        # actually implements a solution
        raise NotImplementedError

    @property
    def settable_params(self):
        # expects dict of parameters that can be set with set_params
        raise NotImplementedError

    @property
    def scoring(self):
        # the default scoring function of this estimator, must be sklearn
        # compatible
        raise NotImplementedError

    @pytest.fixture
    def gp_cls(self):
        raise NotImplementedError

    @pytest.fixture
    def module_cls(self):
        raise NotImplementedError

    @pytest.fixture
    def module_multioutput_cls(self):
        # since multioutput is not currently being tested, not an abstract
        # method
        pass

    @pytest.fixture
    def data(self):
        raise NotImplementedError

    @pytest.fixture
    def gp(self, gp_cls, module_cls, data):
        raise NotImplementedError

    @pytest.fixture
    def gp_fit(self, gp, data):
        X, y = data
        return gp.fit(X, y)

    @pytest.fixture
    def gp_multioutput(self, gp_cls, module_multioutput_cls, data):
        # should be fitted; since it's not currently being tested, not an
        # abstract method
        pass

    @pytest.fixture
    def pipe(self, gp):
        return Pipeline([
            ('noop', None),
            ('gp', gp),
        ])

    ######################
    # saving and loading #
    ######################

    @pytest.mark.xfail(strict=True)
    def test_pickling(self, gp_fit):
        # Currently fails because of issues outside of our control, this test
        # should alert us to when the issue has been fixed. Some issues have
        # been fixed in https://github.com/cornellius-gp/gpytorch/pull/1336 but
        # not all.
        pickle.dumps(gp_fit)

    def test_pickle_error_msg(self, gp_fit):
        # Should eventually be replaced by a test that saves and loads the model
        # using pickle and checks that the predictions are identical
        msg = ("This GPyTorch model cannot be pickled. The reason is probably this:"
               " https://github.com/pytorch/pytorch/issues/38137. "
               "Try using 'dill' instead of 'pickle'.")
        with pytest.raises(pickle.PicklingError, match=msg):
            pickle.dumps(gp_fit)

    def test_deepcopy(self, gp_fit):
        # Should eventually be replaced by a test that saves and loads the model
        # using deepcopy and checks that the predictions are identical
        msg = ("This GPyTorch model cannot be pickled. The reason is probably this:"
               " https://github.com/pytorch/pytorch/issues/38137. "
               "Try using 'dill' instead of 'pickle'.")
        with pytest.raises(pickle.PicklingError, match=msg):
            copy.deepcopy(gp_fit)  # doesn't raise

    def test_clone(self, gp_fit):
        clone(gp_fit)  # doesn't raise

    def test_save_load_params(self, gp_fit, tmpdir):
        gp2 = clone(gp_fit).initialize()

        # check first that parameters are not equal
        for (_, p0), (_, p1) in zip(
                gp_fit.get_all_learnable_params(), gp2.get_all_learnable_params(),
        ):
            assert not (p0 == p1).all()

        # save and load params to gp2
        p_module = tmpdir.join('module.pt')
        p_likelihood = tmpdir.join('likelihood.pt')
        with open(str(p_module), 'wb') as fm, open(str(p_likelihood), 'wb') as fll:
            gp_fit.save_params(f_params=fm, f_likelihood=fll)
        with open(str(p_module), 'rb') as fm, open(str(p_likelihood), 'rb') as fll:
            gp2.load_params(f_params=fm, f_likelihood=fll)

        # now parameters should be equal
        for (n0, p0), (n1, p1) in zip(
                gp_fit.get_all_learnable_params(), gp2.get_all_learnable_params(),
        ):
            assert n0 == n1
            assert_allclose(p0, p1)

    ##############
    # functional #
    ##############

    def test_fit(self, gp_fit, recwarn):
        # fitting does not raise anything and triggers no warning
        assert not recwarn.list

    def test_gp_learns(self, gp_fit):
        history = gp_fit.history
        assert history[0, 'train_loss'] > 0.5 * history[-1, 'train_loss']

    def test_forward(self, gp_fit, data):
        X = data[0]
        y_forward = gp_fit.forward(X)

        for yi in y_forward:
            assert isinstance(yi, torch.distributions.distribution.Distribution)

        total_shape = sum(get_batch_size(p) for p in y_forward)
        assert total_shape == self.n_samples

    def test_predict(self, gp_fit, data):
        X = data[0]
        y_pred = gp_fit.predict(X)

        assert isinstance(y_pred, np.ndarray)
        assert y_pred.shape == (self.n_samples,)
        self.assert_values_differ(y_pred)

    def test_predict_proba(self, gp_fit, data):
        if not self.supports_predict_proba:
            return

        X = data[0]
        y_proba = gp_fit.predict_proba(X)

        assert isinstance(y_proba, np.ndarray)
        assert y_proba.shape == (self.n_samples, self.n_targets)
        self.assert_values_differ(y_proba)

    @pytest.mark.skipif(not torch.cuda.is_available(), reason="no cuda device")
    def test_fit_and_predict_with_cuda(self, gp, data):

        gp.set_params(device='cuda')
        X, y = data
        gp.fit(X, y)
        y_pred = gp.predict(X)
        self.assert_values_differ(y_pred)

    def test_in_sklearn_pipeline(self, pipe, data):
        X, y = data
        # none of this raises an error
        pipe.fit(X, y)
        pipe.predict(X)
        pipe.set_params(**self.settable_params)

    def test_grid_search_works(self, gp, data, recwarn):
        X, y = data
        params = {
            'lr': [0.01, 0.02],
            'max_epochs': [10, 20],
            'likelihood__max_plate_nesting': [1, 2],
        }
        gp.set_params(verbose=0)
        gs = GridSearchCV(gp, params, refit=True, cv=3, scoring=self.scoring)
        gs.fit(X[:60], y[:60])  # for speed

        # sklearn will catch fit failures and raise a warning, we should thus
        # check that no warnings are generated
        assert not recwarn.list

    # Multioutput doesn't work because GPyTorch makes assumptions about the
    # module output that are not compatible with multiple outputs. The tests are
    # left in case this is fixed but they're not being executed.

    @pytest.mark.skip
    def test_fit_multioutput(self, gp_multioutput):
        # doesn't raise
        pass

    @pytest.mark.skip
    def test_multioutput_forward_iter(self, gp_multioutput, data):
        X = data[0]
        y_infer = next(gp_multioutput.forward_iter(X))

        assert isinstance(y_infer, tuple)
        assert len(y_infer) == 3
        assert y_infer[0].shape[0] == min(len(X), gp_multioutput.batch_size)

    @pytest.mark.skip
    def test_multioutput_forward(self, gp_multioutput, data):
        X = data[0]
        y_infer = gp_multioutput.forward(X)

        assert isinstance(y_infer, tuple)
        assert len(y_infer) == 2
        for arr in y_infer:
            assert is_torch_data_type(arr)

        for output in y_infer:
            assert len(output) == self.n_samples

    @pytest.mark.skip
    def test_multioutput_predict(self, gp_multioutput, data):
        X = data[0]

        # does not raise
        y_pred = gp_multioutput.predict(X)

        # Expecting only 1 column containing predict class:
        # (number of samples,)
        assert y_pred.shape == (self.n_samples)
        self.assert_values_differ(y_pred)

    @pytest.mark.skip
    def test_multioutput_predict_proba(self, gp_multioutput, data):
        X = data[0]

        # does not raise
        y_proba = gp_multioutput.predict_proba(X)
        self.assert_values_differ(y_proba)

        # Expecting full output: (number of samples, number of output units)
        assert y_proba.shape == (self.n_samples, self.n_targets)
        # Probabilities, hence these limits
        assert y_proba.min() >= 0
        assert y_proba.max() <= 1

    ##################
    # initialization #
    ##################

    @pytest.mark.parametrize('kwargs,expected', [
        ({}, ""),
        ({
            'likelihood__noise_prior': gpytorch.priors.NormalPrior(0, 1),
            'likelihood__batch_shape': (345,),
        }, ""),
        ({
            'likelihood__noise_prior': gpytorch.priors.NormalPrior(0, 1),
            'optimizer__momentum': 0.567,
        }, ""),
    ])
    def test_set_params_uninitialized_net_correct_message(
            self, gp, kwargs, expected, capsys):
        # When gp is initialized, if module or optimizer need to be
        # re-initialized, alert the user to the fact what parameters
        # were responsible for re-initialization. Note that when the
        # module parameters but not optimizer parameters were changed,
        # the optimizer is re-initialized but not because the
        # optimizer parameters changed.
        gp.set_params(**kwargs)
        msg = capsys.readouterr()[0].strip()
        assert msg == expected

    @pytest.mark.parametrize('kwargs,expected', [
        ({}, ""),
        (
            {'likelihood__max_plate_nesting': 2},
            ("Re-initializing module because the following "
             "parameters were re-set: likelihood__max_plate_nesting.\n"
             "Re-initializing criterion.\n"
             "Re-initializing optimizer.")
        ),
        (
            {
                'likelihood__max_plate_nesting': 2,
                'optimizer__momentum': 0.567,
            },
            ("Re-initializing module because the following "
             "parameters were re-set: likelihood__max_plate_nesting.\n"
             "Re-initializing criterion.\n"
             "Re-initializing optimizer.")
        ),
    ])
    def test_set_params_initialized_net_correct_message(
            self, gp, kwargs, expected, capsys):
        # When gp is initialized, if module or optimizer need to be
        # re-initialized, alert the user to the fact what parameters
        # were responsible for re-initialization. Note that when the
        # module parameters but not optimizer parameters were changed,
        # the optimizer is re-initialized but not because the
        # optimizer parameters changed.
        gp.initialize().set_params(**kwargs)
        msg = capsys.readouterr()[0].strip()
        assert msg == expected

    def test_likelihood_already_initialized_does_not_reinit(self, gp, gp_cls):
        # When the likelihood is already initialized and no params changed, it
        # should just be set as is instead of creating a new instance. In
        # theory, the same should apply to modules but in all the examples here,
        # modules require params, so we cannot test it.

        gp_init = gp.initialize()
        # create a new GP instance using this somewhat convoluted approach
        # because we don't know what arguments are required to initialize from
        # scratch
        params = gp_init.get_params()
        # set likelihood and likelihood to be initialized already
        params['likelihood'] = gp_init.likelihood_
        gp = gp_cls(**params).initialize()

        assert gp.likelihood_ is gp_init.likelihood_

    ##########################
    # probabalistic specific #
    ##########################

    @pytest.mark.parametrize("n_samples", [1, 2, 10])
    def test_sampling(self, gp, data, n_samples):
        X, _ = data
        samples = gp.initialize().sample(X, n_samples=n_samples)
        assert samples.shape == (n_samples, len(X))

        # check that values are not all the same -- this can happen when
        # posterior variances are skipped via a setting
        self.assert_values_differ(samples)

    def test_confidence_region(self, gp_fit, data):
        X, _ = data

        # lower bound should always be lower than upper bound
        lower_1, upper_1 = gp_fit.confidence_region(X, sigmas=1)
        assert (lower_1 < upper_1).all()

        lower_2, upper_2 = gp_fit.confidence_region(X, sigmas=2)
        assert (lower_2 < upper_2).all()

        # higher sigmas -> wider regions
        assert (lower_2 < lower_1).all()
        assert (upper_2 > upper_1).all()

    def test_predict_return_std(self, gp_fit, data):
        if not self.supports_return_std:
            return

        X, _ = data
        y_proba, y_std = gp_fit.predict(X, return_std=True)

        # not a lot we know for sure about the values of the standard deviation,
        # hence only test shape and that they're positive
        assert y_proba.shape == y_std.shape
        assert (y_std > 0).all()
        self.assert_values_differ(y_std)

    def test_predict_return_cov(self, gp_fit, data):
        if not self.supports_return_cov:
            return

        X, _ = data
        msg = ("The 'return_cov' argument is not supported. Please try: "
               "'posterior = next(gpr.forward_iter(X)); posterior.covariance_matrix'.")
        with pytest.raises(NotImplementedError, match=re.escape(msg)):
            gp_fit.predict(X, return_cov=True)


class TestExactGPRegressor(BaseProbabilisticTests):
    """Tests for ExactGPRegressor."""

    ##########################
    # constants and fixtures #
    ##########################

    n_samples = 34
    n_targets = 1
    supports_predict_proba = False
    supports_return_std = True
    supports_return_cov = True
    settable_params = {'gp__likelihood__noise_prior': None}
    scoring = 'neg_mean_squared_error'

    @pytest.fixture
    def data(self):
        X = np.linspace(-8, 8.01, self.n_samples).astype(np.float32)
        y = (np.sin(X) + np.random.randn(len(X)) * 0.2).astype(np.float32)
        return X, y

    @pytest.fixture
    def gp_cls(self):
        from skorch.probabilistic import ExactGPRegressor
        return ExactGPRegressor

    @pytest.fixture
    def module_cls(self):
        return RbfModule

    @pytest.fixture
    def gp(self, gp_cls, module_cls):
        gpr = gp_cls(
            module_cls,
            optimizer=torch.optim.Adam,
            lr=0.1,
            max_epochs=20,
        )
        return gpr

    # pickling and deepcopy work for ExactGPRegressor but not for the others, so
    # override the expected failures here.

    def test_pickling(self, gp_fit):
        # does not raise
        pickle.dumps(gp_fit)

    def test_pickle_error_msg(self, gp_fit):
        # Should eventually be replaced by a test that saves and loads the model
        # using pickle and checks that the predictions are identical
        # FIXME
        pickle.dumps(gp_fit)

    def test_deepcopy(self, gp_fit):
        # FIXME
        copy.deepcopy(gp_fit)  # doesn't raise

    def test_wrong_module_type_raises(self, gp_cls):
        # ExactGPRegressor requires the module to be an ExactGP, if it's not,
        # raise an appropriate error message to the user.
        class VariationalModule(gpytorch.models.ApproximateGP):
            """Defined on root to make it pickleable"""
            def __init__(self, likelihood):
                pass

            def forward(self, x):
                pass

        gp = gp_cls(VariationalModule)
        msg = "ExactGPRegressor requires 'module' to be a gpytorch.models.ExactGP."
        with pytest.raises(TypeError, match=msg):
            gp.initialize()


class TestGPRegressorVariational(BaseProbabilisticTests):
    """Tests for GPRegressor."""

    ##########################
    # constants and fixtures #
    ##########################

    n_samples = 60
    n_targets = 1
    supports_predict_proba = False
    supports_return_std = True
    supports_return_cov = True
    settable_params = {'gp__module__eps': 1e-5}
    scoring = 'neg_mean_squared_error'

    @pytest.fixture
    def data(self):
        X = np.linspace(-8, 8.01, self.n_samples).astype(np.float32)
        y = (np.sin(X) + np.random.randn(len(X)) * 0.2).astype(np.float32)
        return X, y

    @pytest.fixture
    def gp_cls(self):
        from skorch.probabilistic import GPRegressor
        return GPRegressor

    @pytest.fixture
    def module_cls(self):
        return VariationalRegressionModule

    @pytest.fixture
    def gp(self, gp_cls, module_cls, data):
        X, y = data
        gpr = gp_cls(
            module_cls,
            module__inducing_points=torch.from_numpy(X[:10]),

            criterion=gpytorch.mlls.VariationalELBO,
            criterion__num_data=int(0.8 * len(y)),
            batch_size=24,
        )
        # we want to make sure batching is properly tested
        assert gpr.batch_size < self.n_samples
        return gpr


class TestGPBinaryClassifier(BaseProbabilisticTests):
    """Tests for GPBinaryClassifier."""

    ##########################
    # constants and fixtures #
    ##########################

    n_samples = 50
    n_targets = 2
    supports_predict_proba = True
    supports_return_std = False
    supports_return_cov = False
    settable_params = {'gp__module__eps': 1e-5}
    scoring = 'neg_mean_squared_error'

    @pytest.fixture
    def data(self):
        X = np.linspace(-8, 8.01, self.n_samples).astype(np.float32)
        y = (np.sin(X) + np.random.randn(len(X)) * 0.2 > 0).astype(np.int64)
        return X, y

    @pytest.fixture
    def gp_cls(self):
        from skorch.probabilistic import GPBinaryClassifier
        return GPBinaryClassifier

    @pytest.fixture
    def module_cls(self):
        return VariationalBinaryClassificationModule

    @pytest.fixture
    def gp(self, gp_cls, module_cls, data):
        X, y = data
        gpc = gp_cls(
            module_cls,
            module__inducing_points=torch.from_numpy(X[:10]),

            criterion=gpytorch.mlls.VariationalELBO,
            criterion__num_data=int(0.8 * len(y)),
            batch_size=24,
        )
        # we want to make sure batching is properly tested
        assert gpc.batch_size < self.n_samples
        return gpc



"""Tests for callbacks/logging.py"""

from functools import partial
import os
import tempfile
import unittest.mock
from unittest.mock import Mock
from unittest.mock import call, patch

import numpy as np
import pytest
import torch
from torch import nn

from skorch.tests.conftest import neptune_installed
from skorch.tests.conftest import sacred_installed
from skorch.tests.conftest import wandb_installed
from skorch.tests.conftest import tensorboard_installed
from skorch.tests.conftest import mlflow_installed

@pytest.mark.skipif(
    not neptune_installed, reason='neptune is not installed')
class TestNeptune:
    # fields logged by on_train_begin and on_train_end
    NUM_BASE_METRICS = 9

    @pytest.fixture
    def net_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.fixture
    def data(self, classifier_data):
        X, y = classifier_data
        # accelerate training since we don't care for the loss
        X, y = X[:40], y[:40]
        return X, y

    @pytest.fixture
    def neptune_logger_cls(self):
        from skorch.callbacks import NeptuneLogger
        return NeptuneLogger

    @pytest.fixture
    def neptune_run_object(self):
        try:
            # neptune-client=0.9.0+ package structure
            import neptune.new as neptune
        except ImportError:
            # neptune-client>=1.0.0 package structure
            import neptune

        run = neptune.init_run(
            project="tests/dry-run",
            mode="offline",
        )
        return run

    @pytest.fixture
    def mock_experiment(self, neptune_run_object):
        with neptune_run_object as run:
            return unittest.mock.create_autospec(run)

    @pytest.fixture
    def net_fitted(
            self,
            net_cls,
            classifier_module,
            data,
            neptune_logger_cls,
            mock_experiment,
    ):
        return net_cls(
            classifier_module,
            callbacks=[neptune_logger_cls(mock_experiment)],
            max_epochs=3,
        ).fit(*data)

    def test_experiment_closed_automatically(self, net_fitted, mock_experiment):
        assert mock_experiment.stop.call_count == 1

    def test_experiment_log_call_counts(self, net_fitted, mock_experiment):
        # (3 x dur + 3 x train_loss + 3 x valid_loss + 3 x valid_acc = 12) + base metrics
        assert mock_experiment.__getitem__.call_count == 12 + self.NUM_BASE_METRICS

    def test_experiment_not_closed(
            self,
            net_cls,
            classifier_module,
            data,
            neptune_logger_cls,
            mock_experiment,
    ):
        net_cls(
            classifier_module,
            callbacks=[
                neptune_logger_cls(mock_experiment, close_after_train=False)],
            max_epochs=2,
        ).fit(*data)
        assert mock_experiment.stop.call_count == 0

    def test_ignore_keys(
            self,
            net_cls,
            classifier_module,
            data,
            neptune_logger_cls,
            mock_experiment,
    ):
        # ignore 'dur' and 'valid_loss', 'unknown' doesn't exist but
        # this should not cause a problem
        npt = neptune_logger_cls(
            mock_experiment, keys_ignored=['dur', 'valid_loss', 'unknown'])
        net_cls(
            classifier_module,
            callbacks=[npt],
            max_epochs=3,
        ).fit(*data)

        # (3 epochs x 2 epoch metrics = 6 calls) + base metrics
        assert mock_experiment.__getitem__.call_count == 6 + self.NUM_BASE_METRICS

    def test_keys_ignored_is_string(self, neptune_logger_cls, mock_experiment):
        npt = neptune_logger_cls(
            mock_experiment, keys_ignored='a-key').initialize()
        expected = {'a-key', 'batches'}
        assert npt.keys_ignored_ == expected

    def test_fit_with_real_experiment(
            self,
            net_cls,
            classifier_module,
            data,
            neptune_logger_cls,
            neptune_run_object,
    ):
        net = net_cls(
            classifier_module,
            callbacks=[neptune_logger_cls(neptune_run_object)],
            max_epochs=5,
        )
        net.fit(*data)

        assert neptune_run_object.exists('training/epoch_duration')
        assert neptune_run_object.exists('training/train/epoch/loss')
        assert neptune_run_object.exists('training/validation/epoch/loss')
        assert neptune_run_object.exists('training/validation/epoch/acc')

        # Checkpoint callback was not used
        assert not neptune_run_object.exists('training/model/checkpoint')

    def test_log_on_batch_level_on(
            self,
            net_cls,
            classifier_module,
            data,
            neptune_logger_cls,
            mock_experiment,
    ):
        net = net_cls(
            classifier_module,
            callbacks=[neptune_logger_cls(mock_experiment, log_on_batch_end=True)],
            max_epochs=5,
            batch_size=4,
            train_split=False
        )
        net.fit(*data)

        # (5 epochs x (40/4 batches x 2 batch metrics + 2 epoch metrics) = 110 calls) + base metrics
        assert mock_experiment.__getitem__.call_count == 110 + self.NUM_BASE_METRICS
        mock_experiment['training']['train']['batch']['batch_size'].log.assert_any_call(4)

    def test_log_on_batch_level_off(
            self,
            net_cls,
            classifier_module,
            data,
            neptune_logger_cls,
            mock_experiment,
    ):
        net = net_cls(
            classifier_module,
            callbacks=[neptune_logger_cls(mock_experiment, log_on_batch_end=False)],
            max_epochs=5,
            batch_size=4,
            train_split=False
        )
        net.fit(*data)

        # (5 epochs x 2 epoch metrics = 10 calls) + base metrics
        assert mock_experiment.__getitem__.call_count == 10 + self.NUM_BASE_METRICS

        call_args = mock_experiment['training']['train'].__getitem__.call_args_list
        assert call('epoch') in call_args
        assert call('batch') not in call_args

        call_args = mock_experiment['training']['validation'].__getitem__.call_args_list
        assert call('epoch') in call_args
        assert call('batch') not in call_args

    def test_fit_with_real_experiment_saving_checkpoints(
            self,
            net_cls,
            classifier_module,
            data,
            neptune_logger_cls,
            neptune_run_object,
    ):
        try:
            # neptune-client=0.9.0+ package structure
            from neptune.new.attributes.file_set import FileSet
        except ImportError:
            # neptune-client>=1.0.0 package structure
            from neptune.attributes.file_set import FileSet
        from skorch.callbacks import Checkpoint

        with tempfile.TemporaryDirectory() as directory:
            net = net_cls(
                classifier_module,
                callbacks=[
                    neptune_logger_cls(
                        run=neptune_run_object,
                        close_after_train=False,
                    ),
                    Checkpoint(dirname=directory),
                ],
                max_epochs=5,
            )
            net.fit(*data)
            neptune_run_object['training/model/checkpoint'].upload_files(directory)

        assert neptune_run_object.exists('training/train/epoch/loss')
        assert neptune_run_object.exists('training/validation/epoch/loss')
        assert neptune_run_object.exists('training/validation/epoch/acc')

        assert neptune_run_object.exists('training/model/checkpoint')
        assert isinstance(
            neptune_run_object.get_structure()['training']['model']['checkpoint'],
            FileSet,
        )

        neptune_run_object.stop()

@pytest.mark.skipif(
    not sacred_installed, reason='Sacred is not installed')
class TestSacred:
    @pytest.fixture
    def net_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.fixture
    def data(self, classifier_data):
        X, y = classifier_data
        # accelerate training since we don't care for the loss
        X, y = X[:40], y[:40]
        return X, y

    @pytest.fixture
    def sacred_logger_cls(self):
        from skorch.callbacks import SacredLogger
        return SacredLogger

    @pytest.fixture
    def sacred_experiment_cls(self):
        from sacred import Experiment
        return Experiment

    @pytest.fixture
    def mock_experiment(self, sacred_experiment_cls):
        mock = Mock(spec=sacred_experiment_cls)
        mock.log_scalar = Mock()
        return mock

    @pytest.fixture
    def net_fitted(
            self,
            net_cls,
            classifier_module,
            data,
            sacred_logger_cls,
            mock_experiment,
    ):
        return net_cls(
            classifier_module,
            callbacks=[sacred_logger_cls(mock_experiment)],
            max_epochs=3,
        ).fit(*data)

    def test_ignore_keys(
            self,
            net_cls,
            classifier_module,
            data,
            sacred_logger_cls,
            mock_experiment,
    ):
        # ignore 'dur' and 'valid_loss', 'unknown' doesn't exist but
        # this should not cause a problem
        logger = sacred_logger_cls(
            mock_experiment, keys_ignored=['dur', 'valid_loss', 'unknown'])
        net_cls(
            classifier_module,
            callbacks=[logger],
            max_epochs=3,
        ).fit(*data)

        # 3 epochs x 2 epoch metrics = 6 calls
        assert mock_experiment.log_scalar.call_count == 6
        call_args = [args[0][0] for args in mock_experiment.log_scalar.call_args_list]
        assert 'valid_loss' not in call_args

    def test_keys_ignored_is_string(self, sacred_logger_cls, mock_experiment):
        logger = sacred_logger_cls(
            mock_experiment, keys_ignored='a-key').initialize()
        expected = {'a-key', 'batches'}
        assert logger.keys_ignored_ == expected

    def test_fit_with_real_experiment(
            self,
            net_cls,
            classifier_module,
            data,
            sacred_logger_cls,
            sacred_experiment_cls,
    ):
        experiment = sacred_experiment_cls()

        @experiment.main
        def experiment_main(_run):
            net = net_cls(
                classifier_module,
                callbacks=[sacred_logger_cls(_run)],
                max_epochs=5,
            )
            net.fit(*data)

        experiment.run()

    def test_log_on_batch_level_on(
            self,
            net_cls,
            classifier_module,
            data,
            sacred_logger_cls,
            mock_experiment,
    ):
        net = net_cls(
            classifier_module,
            callbacks=[sacred_logger_cls(mock_experiment, log_on_batch_end=True)],
            max_epochs=5,
            batch_size=4,
            train_split=False
        )
        net.fit(*data)

        # 5 epochs x (40/4 batches x 2 batch metrics + 2 epoch metrics) = 110 calls
        assert mock_experiment.log_scalar.call_count == 110
        mock_experiment.log_scalar.assert_any_call('train_batch_size_batch', 4)

        logged_keys = [
            call_args.args[0] for call_args in mock_experiment.log_scalar.call_args_list
        ]
        # This is a batch-only metric.
        assert 'train_batch_size_epoch' not in logged_keys

    def test_log_on_batch_level_off(
            self,
            net_cls,
            classifier_module,
            data,
            sacred_logger_cls,
            mock_experiment,
    ):
        net = net_cls(
            classifier_module,
            callbacks=[sacred_logger_cls(mock_experiment, log_on_batch_end=False)],
            max_epochs=5,
            batch_size=4,
            train_split=False
        )
        net.fit(*data)

        # 5 epochs x 2 epoch metrics = 10 calls
        assert mock_experiment.log_scalar.call_count == 10
        call_args_list = mock_experiment.log_scalar.call_args_list
        assert call('train_batch_size_batch', 4) not in call_args_list

@pytest.mark.skipif(
    not wandb_installed, reason='wandb is not installed')
class TestWandb:
    @pytest.fixture
    def net_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.fixture
    def data(self, classifier_data):
        X, y = classifier_data
        # accelerate training since we don't care for the loss
        X, y = X[:40], y[:40]
        return X, y

    @pytest.fixture
    def wandb_logger_cls(self):
        from skorch.callbacks import WandbLogger
        return WandbLogger

    @pytest.fixture
    def wandb_run_cls(self):
        import wandb
        os.environ['WANDB_MODE'] = 'dryrun' # run offline
        wandb_version = tuple(map(int, wandb.__version__.split('.')[:2]))
        if wandb_version >= (0, 10):
            return wandb.init(anonymous="allow")
        else:
            with wandb.init(anonymous="allow") as run:
                return run

    @pytest.fixture
    def mock_run(self):
        mock = Mock()
        mock.log = Mock()
        mock.watch = Mock()
        mock.dir = '.'
        return mock

    def test_ignore_keys(
            self,
            net_cls,
            classifier_module,
            data,
            wandb_logger_cls,
            mock_run,
    ):
        # ignore 'dur' and 'valid_loss', 'unknown' doesn't exist but
        # this should not cause a problem
        wandb_cb = wandb_logger_cls(
            mock_run, keys_ignored=['dur', 'valid_loss', 'unknown'])
        net_cls(
            classifier_module,
            callbacks=[wandb_cb],
            max_epochs=3,
        ).fit(*data)

        # 3 epochs = 3 calls
        assert mock_run.log.call_count == 3
        assert mock_run.watch.call_count == 1
        call_args = [args[0][0] for args in mock_run.log.call_args_list]
        assert 'valid_loss' not in call_args

    def test_keys_ignored_is_string(self, wandb_logger_cls, mock_run):
        wandb_cb = wandb_logger_cls(
            mock_run, keys_ignored='a-key').initialize()
        expected = {'a-key', 'batches'}
        assert wandb_cb.keys_ignored_ == expected

    def test_fit_with_real_experiment(
            self,
            net_cls,
            classifier_module,
            data,
            wandb_logger_cls,
            wandb_run_cls,
    ):
        net = net_cls(
            classifier_module,
            callbacks=[wandb_logger_cls(wandb_run_cls)],
            max_epochs=5,
        )
        net.fit(*data)

class TestPrintLog:
    @pytest.fixture
    def print_log_cls(self):
        from skorch.callbacks import PrintLog
        keys_ignored = ['dur', 'event_odd']
        return partial(PrintLog, sink=Mock(), keys_ignored=keys_ignored)

    @pytest.fixture
    def print_log(self, print_log_cls):
        return print_log_cls().initialize()

    @pytest.fixture
    def scoring_cls(self):
        from skorch.callbacks import EpochScoring
        return EpochScoring

    @pytest.fixture
    def mse_scoring(self, scoring_cls):
        return scoring_cls(
            'neg_mean_squared_error',
            name='nmse',
        ).initialize()

    @pytest.fixture
    def odd_epoch_callback(self):
        from skorch.callbacks import Callback
        class OddEpochCallback(Callback):
            def on_epoch_end(self, net, **kwargs):
                net.history[-1]['event_odd'] = bool(len(net.history) % 2)

        return OddEpochCallback().initialize()

    @pytest.fixture
    def net(self, net_cls, module_cls, train_split, mse_scoring,
            odd_epoch_callback, print_log, data):
        net = net_cls(
            module_cls, batch_size=1, train_split=train_split,
            callbacks=[mse_scoring, odd_epoch_callback], max_epochs=2)
        net.initialize()
        # replace default PrintLog with test PrintLog
        net.callbacks_[-1] = ('print_log', print_log)
        return net.partial_fit(*data)

    @pytest.fixture
    def history(self, net):
        return net.history

    # pylint: disable=unused-argument
    @pytest.fixture
    def sink(self, history, print_log):
        # note: the history fixture is required even if not used because it
        # triggers the calls on print_log
        return print_log.sink

    @pytest.fixture
    def ansi(self):
        from skorch.utils import Ansi
        return Ansi

    def test_call_count(self, sink):
        # header + lines + 2 epochs
        assert sink.call_count == 4

    def test_header(self, sink):
        header = sink.call_args_list[0][0][0]
        columns = header.split()
        expected = ['epoch', 'nmse', 'train_loss', 'valid_loss']
        assert columns == expected

    def test_lines(self, sink):
        lines = sink.call_args_list[1][0][0].split()
        # Lines have length 2 + length of column, or 8 if the column
        # name is short and the values are floats.
        expected = [
            '-' * (len('epoch') + 2),
            '-' * 8,
            '-' * (len('train_loss') + 2),
            '-' * (len('valid_loss') + 2),
        ]
        assert lines
        assert lines == expected

    @pytest.mark.parametrize('epoch', [0, 1])
    def test_first_row(self, sink, ansi, epoch, history):
        row = sink.call_args_list[epoch + 2][0][0]
        items = row.split()

        # epoch, nmse, valid, train
        assert len(items) == 4

        # epoch, starts at 1
        assert items[0] == str(epoch + 1)

        # is best
        are_best = [
            history[epoch, 'nmse_best'],
            history[epoch, 'train_loss_best'],
            history[epoch, 'valid_loss_best'],
        ]

        # test that cycled colors are used if best
        for item, color, is_best in zip(items[1:], list(ansi)[1:], are_best):
            if is_best:
                # if best, text colored
                assert item.startswith(color.value)
                assert item.endswith(ansi.ENDC.value)
            else:
                # if not best, text is only float, so converting possible
                float(item)

    def test_args_passed_to_tabulate(self, history):
        with patch('skorch.callbacks.logging.tabulate') as tab:
            from skorch.callbacks import PrintLog
            print_log = PrintLog(
                tablefmt='latex',
                floatfmt='.9f',
            ).initialize()
            print_log.table(history[-1])

            assert tab.call_count == 1
            assert tab.call_args_list[0][1]['tablefmt'] == 'latex'
            assert tab.call_args_list[0][1]['floatfmt'] == '.9f'

    def test_with_additional_key(self, history, print_log_cls):
        keys_ignored = ['event_odd']  # 'dur' no longer ignored
        print_log = print_log_cls(
            sink=Mock(), keys_ignored=keys_ignored).initialize()
        # does not raise
        print_log.on_epoch_end(Mock(history=history))

        header = print_log.sink.call_args_list[0][0][0]
        columns = header.split()
        expected = ['epoch', 'nmse', 'train_loss', 'valid_loss', 'dur']
        assert columns == expected

    def test_keys_ignored_as_str(self, print_log_cls):
        print_log = print_log_cls(keys_ignored='a-key').initialize()
        assert print_log.keys_ignored_ == {'a-key', 'batches'}

        print_log.initialize()
        assert print_log.keys_ignored_ == set(['a-key', 'batches'])

    def test_keys_ignored_is_None(self, print_log_cls):
        print_log = print_log_cls(keys_ignored=None)
        assert print_log.keys_ignored is None

        print_log.initialize()
        assert print_log.keys_ignored_ == set(['batches'])

    def test_with_event_key(self, history, print_log_cls):
        print_log = print_log_cls(sink=Mock(), keys_ignored=None).initialize()
        # history has two epochs, write them one by one
        print_log.on_epoch_end(Mock(history=history[:-1]))
        print_log.on_epoch_end(Mock(history=history))

        header = print_log.sink.call_args_list[0][0][0]
        columns = header.split()
        expected = ['epoch', 'nmse', 'train_loss', 'valid_loss', 'odd', 'dur']
        assert columns == expected

        odd_row = print_log.sink.call_args_list[2][0][0].split()
        even_row = print_log.sink.call_args_list[3][0][0].split()
        assert len(odd_row) == 6  # odd row has entries in every column
        assert odd_row[4] == '+'  # including '+' sign for the 'event_odd'
        assert len(even_row) == 5  # even row does not have 'event_odd' entry

    def test_witout_valid_data(
            self, net_cls, module_cls, mse_scoring, print_log, data):
        net = net_cls(
            module_cls, batch_size=1, train_split=None,
            callbacks=[mse_scoring], max_epochs=2)
        net.initialize()
        # replace default PrintLog with test PrintLog
        net.callbacks_[-1] = ('print_log', print_log)
        net.partial_fit(*data)

        sink = print_log.sink
        row = sink.call_args_list[2][0][0]
        items = row.split()

        assert len(items) == 2  # no valid, only epoch and train

    def test_print_not_skipped_if_verbose(self, capsys):
        from skorch.callbacks import PrintLog

        print_log = PrintLog().initialize()
        net = Mock(history=[{'loss': 123}], verbose=1)

        print_log.on_epoch_end(net)

        stdout = capsys.readouterr()[0]
        result = [x.strip() for x in stdout.split()]
        expected = ['loss', '------', '123']
        assert result == expected

    def test_print_skipped_if_not_verbose(self, capsys):
        from skorch.callbacks import PrintLog

        print_log = PrintLog().initialize()
        net = Mock(history=[{'loss': 123}], verbose=0)

        print_log.on_epoch_end(net)

        stdout = capsys.readouterr()[0]
        assert not stdout


class TestProgressBar:
    @pytest.fixture
    def progressbar_cls(self):
        from skorch.callbacks import ProgressBar
        return ProgressBar

    @pytest.fixture
    def net_cls(self):
        """very simple network that trains for 2 epochs"""
        from skorch import NeuralNetRegressor
        from skorch.toy import make_regressor

        module_cls = make_regressor(
            input_units=1,
            num_hidden=0,
            output_units=1,
        )

        return partial(
            NeuralNetRegressor,
            module=module_cls,
            train_split=None,
            max_epochs=2,
            batch_size=10)

    @pytest.fixture(scope='module')
    def data(self):
        X = np.zeros((20, 1), dtype='float32')
        y = np.zeros((20, 1), dtype='float32')
        return X, y

    @pytest.mark.parametrize('postfix', [
        [],
        ['train_loss'],
        ['train_loss', 'valid_loss'],
        ['doesnotexist'],
        ['train_loss', 'doesnotexist'],
    ])
    def test_invalid_postfix(self, postfix, net_cls, progressbar_cls, data):
        net = net_cls(callbacks=[
            progressbar_cls(postfix_keys=postfix),
        ])
        net.fit(*data)

    @patch('tqdm.tqdm')
    @pytest.mark.parametrize('scheme,expected_total', [
        ('auto', [2, 2]),
        ('count', [None, 2]),
        (None, [None, None]),
        (2, [2, 2]),  # correct number of batches_per_epoch (20 // 10)
        (3, [3, 3]),  # offset by +1, should still work
        (1, [1, 1]),  # offset by -1, should still work
    ])
    def test_different_count_schemes(
            self, tqdm_mock, scheme, expected_total, net_cls, progressbar_cls, data):
        net = net_cls(callbacks=[
            progressbar_cls(batches_per_epoch=scheme),
        ])
        net.fit(*data)
        assert tqdm_mock.call_count == 2
        for i, total in enumerate(expected_total):
            assert tqdm_mock.call_args_list[i][1]['total'] == total

    def test_pickle(self, net_cls, progressbar_cls, data):
        # pickling was an issue since TQDM progress bar instances cannot
        # be pickled. Test pickling and restoration.
        import pickle

        net = net_cls(callbacks=[
            progressbar_cls(),
        ])
        net.fit(*data)
        dump = pickle.dumps(net)

        net = pickle.loads(dump)
        net.fit(*data)


@pytest.mark.skipif(
    not tensorboard_installed, reason='tensorboard is not installed')
class TestTensorBoard:
    @pytest.fixture
    def net_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.fixture
    def data(self, classifier_data):
        X, y = classifier_data
        # accelerate training since we don't care for the loss
        X, y = X[:40], y[:40]
        return X, y

    @pytest.fixture
    def tensorboard_cls(self):
        from skorch.callbacks import TensorBoard
        return TensorBoard

    @pytest.fixture
    def summary_writer_cls(self):
        from torch.utils.tensorboard import SummaryWriter
        return SummaryWriter

    @pytest.fixture
    def mock_writer(self, summary_writer_cls):
        mock = Mock(spec=summary_writer_cls)
        return mock

    @pytest.fixture
    def net_fitted(
            self,
            net_cls,
            classifier_module,
            data,
            tensorboard_cls,
            mock_writer,
    ):
        return net_cls(
            classifier_module,
            callbacks=[tensorboard_cls(mock_writer)],
            max_epochs=3,
        ).fit(*data)

    @pytest.mark.skipif(
        True, reason="Waiting for proper implementation of graph tracing")
    def test_graph_added_once(self, net_fitted, mock_writer):
        # graph should just be added once
        assert mock_writer.add_graph.call_count == 1

    @pytest.mark.skipif(
        True, reason="Waiting for proper implementation of graph tracing")
    def test_include_graph_false(
            self,
            net_cls,
            classifier_module,
            data,
            tensorboard_cls,
            mock_writer,
    ):
        net_cls(
            classifier_module,
            callbacks=[tensorboard_cls(mock_writer, include_graph=False)],
            max_epochs=2,
        ).fit(*data)
        assert mock_writer.add_graph.call_count == 0

    def test_writer_closed_automatically(self, net_fitted, mock_writer):
        assert mock_writer.close.call_count == 1

    def test_writer_not_closed(
            self,
            net_cls,
            classifier_module,
            data,
            tensorboard_cls,
            mock_writer,
    ):
        net_cls(
            classifier_module,
            callbacks=[tensorboard_cls(mock_writer, close_after_train=False)],
            max_epochs=2,
        ).fit(*data)
        assert mock_writer.close.call_count == 0

    def test_keys_from_history_logged(self, net_fitted, mock_writer):
        add_scalar = mock_writer.add_scalar

        # 3 epochs with 4 keys
        assert add_scalar.call_count == 3 * 4
        keys = {call_args[1]['tag'] for call_args in add_scalar.call_args_list}
        expected = {'dur', 'Loss/train_loss', 'Loss/valid_loss', 'Loss/valid_acc'}
        assert keys == expected

    def test_ignore_keys(
            self,
            net_cls,
            classifier_module,
            data,
            tensorboard_cls,
            mock_writer,
    ):
        # ignore 'dur' and 'valid_loss', 'unknown' doesn't exist but
        # this should not cause a problem
        tb = tensorboard_cls(
            mock_writer, keys_ignored=['dur', 'valid_loss', 'unknown'])
        net_cls(
            classifier_module,
            callbacks=[tb],
            max_epochs=3,
        ).fit(*data)
        add_scalar = mock_writer.add_scalar

        keys = {call_args[1]['tag'] for call_args in add_scalar.call_args_list}
        expected = {'Loss/train_loss', 'Loss/valid_acc'}
        assert keys == expected

    def test_keys_ignored_is_string(self, tensorboard_cls, mock_writer):
        tb = tensorboard_cls(mock_writer, keys_ignored='a-key').initialize()
        expected = {'a-key', 'batches'}
        assert tb.keys_ignored_ == expected

    def test_other_key_mapper(
            self,
            net_cls,
            classifier_module,
            data,
            tensorboard_cls,
            mock_writer,
    ):
        # just map all keys to uppercase
        tb = tensorboard_cls(mock_writer, key_mapper=lambda s: s.upper())
        net_cls(
            classifier_module,
            callbacks=[tb],
            max_epochs=3,
        ).fit(*data)
        add_scalar = mock_writer.add_scalar

        keys = {call_args[1]['tag'] for call_args in add_scalar.call_args_list}
        expected = {'DUR', 'TRAIN_LOSS', 'VALID_LOSS', 'VALID_ACC'}
        assert keys == expected

    @pytest.fixture
    def add_scalar_maybe(self, tensorboard_cls, mock_writer):
        tb = tensorboard_cls(mock_writer)
        return tb.add_scalar_maybe

    @pytest.fixture
    def history(self):
        return [
            {'loss': 0.1, 'epoch': 1, 'foo': ['invalid', 'type']},
            {'loss': 0.2, 'epoch': 2, 'foo': ['invalid', 'type']},
        ]

    def test_add_scalar_maybe_uses_last_epoch_values(
            self, add_scalar_maybe, mock_writer, history):
        add_scalar_maybe(history, key='loss', tag='myloss', global_step=2)
        call_kwargs = mock_writer.add_scalar.call_args_list[0][1]
        assert call_kwargs['tag'] == 'myloss'
        assert call_kwargs['scalar_value'] == 0.2
        assert call_kwargs['global_step'] == 2

    def test_add_scalar_maybe_infers_epoch(
            self, add_scalar_maybe, mock_writer, history):
        # don't indicate 'global_step' value
        add_scalar_maybe(history, key='loss', tag='myloss')
        call_kwargs = mock_writer.add_scalar.call_args_list[0][1]
        assert call_kwargs['global_step'] == 2

    def test_add_scalar_maybe_unknown_key_does_not_raise(
            self, tensorboard_cls, summary_writer_cls, history):
        tb = tensorboard_cls(summary_writer_cls())
        # does not raise:
        tb.add_scalar_maybe(history, key='unknown', tag='bar')

    def test_add_scalar_maybe_wrong_type_does_not_raise(
            self, tensorboard_cls, summary_writer_cls, history):
        tb = tensorboard_cls(summary_writer_cls())
        # value of 'foo' is a list but that does not raise:
        tb.add_scalar_maybe(history, key='foo', tag='bar')

    def test_fit_with_real_summary_writer(
            self,
            net_cls,
            classifier_module,
            data,
            tensorboard_cls,
            summary_writer_cls,
            tmp_path,
    ):
        path = str(tmp_path)

        net = net_cls(
            classifier_module,
            callbacks=[tensorboard_cls(summary_writer_cls(path))],
            max_epochs=5,
        )
        net.fit(*data)

        # is not empty
        assert os.listdir(path)

    def test_fit_with_dict_input(
            self,
            net_cls,
            classifier_module,
            data,
            tensorboard_cls,
            summary_writer_cls,
            tmp_path,
    ):
        from skorch.toy import MLPModule
        path = str(tmp_path)
        X, y = data

        # create a dictionary with unordered keys
        X_dict = {k: X[:, i:i + 4] for k, i in zip('cebad', range(0, X.shape[1], 4))}

        class MyModule(MLPModule):
            # use different order for args here
            def forward(self, b, e, c, d, a, **kwargs):
                X = torch.cat((b, e, c, d, a), 1)
                return super().forward(X, **kwargs)

        net = net_cls(
            MyModule(output_nonlin=nn.Softmax(dim=-1)),
            callbacks=[tensorboard_cls(summary_writer_cls(path))],
            max_epochs=5,
        )
        net.fit(X_dict, y)

        # is not empty
        assert os.listdir(path)


@pytest.mark.skipif(
    not mlflow_installed, reason='mlflow is not installed')
class TestMLflowLogger:
    @pytest.fixture
    def net_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.fixture
    def logger_cls(self):
        from skorch.callbacks import MlflowLogger
        return MlflowLogger

    @pytest.fixture
    def data(self, classifier_data):
        X, y = classifier_data
        X, y = X[:40], y[:40]
        return X, y

    @pytest.fixture
    def mock_run(self):
        from mlflow.entities import Run
        return Mock(Run)

    @pytest.fixture
    def mock_client(self):
        from mlflow.tracking import MlflowClient
        return Mock(MlflowClient)

    @pytest.fixture
    def logger_mock_cls(self, logger_cls, mock_run, mock_client):
        return partial(logger_cls, mock_run, mock_client)

    @pytest.fixture
    def net_builder_cls(self, net_cls, classifier_module, data):
        def builder(*args, **kwargs):
            return net_cls(classifier_module, *args, **kwargs).fit(*data)
        return builder

    @pytest.fixture
    def net_fitted(self, logger_mock_cls, net_builder_cls):
        return net_builder_cls(callbacks=[logger_mock_cls()], max_epochs=3)

    def test_run_default(self, monkeypatch, logger_cls, mock_run, mock_client):
        import mlflow
        mock_active_run = Mock(mlflow.active_run, return_value=mock_run)
        monkeypatch.setattr(mlflow, 'active_run', mock_active_run)
        logger = logger_cls(client=mock_client).initialize()
        assert mock_active_run.called
        assert logger.run_ == mock_run

    def test_client_default(self, monkeypatch, logger_cls, mock_run, mock_client):
        import mlflow.tracking
        monkeypatch.setattr(mlflow.tracking, 'MlflowClient', mock_client)
        logger = logger_cls(run=mock_run).initialize()
        assert mock_client.called
        assert logger.client_ == mock_client()

    def test_keys_from_history_logged(self, net_fitted, mock_client):
        assert mock_client.log_metric.call_count == 3 * 4
        keys = {call_args[0][1] for call_args in mock_client.log_metric.call_args_list}
        expected = {'dur', 'train_loss', 'valid_loss', 'valid_acc'}
        assert keys == expected

    def test_ignore_keys(self, logger_mock_cls, net_builder_cls):
        # ignore 'dur' and 'valid_loss', 'unknown' doesn't exist but
        # this should not cause a problem
        logger = logger_mock_cls(keys_ignored=['dur', 'valid_loss', 'unknown'])
        net_builder_cls(callbacks=[logger], max_epochs=3)
        keys = {
            call_args[0][1]
            for call_args in logger.client_.log_metric.call_args_list
        }
        expected = {'train_loss', 'valid_acc'}
        assert keys == expected

    def test_keys_ignored_is_string(self, logger_mock_cls):
        logger = logger_mock_cls(keys_ignored='a-key').initialize()
        expected = {'a-key', 'batches'}
        assert logger.keys_ignored_ == expected

    @pytest.mark.parametrize(
        'log_on_batch_end, log_on_epoch_end, batch_suffix, epoch_suffix',
        [(False, False, '', ''),
         (True, False, '', ''),
         (False, True, '', ''),
         (True, True, '_batch', '_epoch')]
    )
    def test_epoch_batch_suffixes_defaults(
            self,
            logger_mock_cls,
            log_on_batch_end,
            log_on_epoch_end,
            batch_suffix,
            epoch_suffix,
    ):
        logger = logger_mock_cls(
            log_on_batch_end=log_on_batch_end,
            log_on_epoch_end=log_on_epoch_end
        ).initialize()
        assert logger.batch_suffix_ == batch_suffix
        assert logger.epoch_suffix_ == epoch_suffix

    @pytest.mark.parametrize('batch_suffix', ['', '_foo'])
    @pytest.mark.parametrize('epoch_suffix', ['', '_bar'])
    def test_epoch_batch_custom_suffix(
            self,
            logger_mock_cls,
            batch_suffix,
            epoch_suffix,
    ):
        logger = logger_mock_cls(
            log_on_batch_end=True,
            log_on_epoch_end=True,
            batch_suffix=batch_suffix,
            epoch_suffix=epoch_suffix,
        ).initialize()
        assert logger.batch_suffix_ == batch_suffix
        assert logger.epoch_suffix_ == epoch_suffix

    def test_dont_log_epoch_metrics(self, logger_mock_cls, net_builder_cls):
        logger = logger_mock_cls(
            log_on_batch_end=True,
            log_on_epoch_end=False,
            batch_suffix='_batch',
            epoch_suffix='_epoch',
        )
        net_builder_cls(batch_size=10, callbacks=[logger], max_epochs=3)
        assert all(
            call[0][1].endswith('_batch')
            for call in logger.client_.log_metric.call_args_list
        )

    def test_log_epochs_with_step(self, net_fitted, mock_client):
        expected = [x for x in range(1, 4) for _ in range(4)]
        actual = [call[1].get('step') for call in mock_client.log_metric.call_args_list]
        assert expected == actual

    def test_log_batch_with_step(self, logger_mock_cls, net_builder_cls):
        logger = logger_mock_cls(log_on_batch_end=True, log_on_epoch_end=False)
        net_builder_cls(batch_size=10, callbacks=[logger], max_epochs=4)
        expected = [x for x in range(1, 21) for _ in range(2)]
        actual = [
            call[1].get('step')
            for call in logger.client_.log_metric.call_args_list
        ]
        assert expected == actual

    def test_artifact_filenames(self, net_fitted, mock_client):
        keys = {call_args[0][1].name
                for call_args in mock_client.log_artifact.call_args_list}
        expected = {'params.pth', 'optimizer.pth', 'criterion.pth', 'history.json'}
        assert keys == expected

    def test_artifact_in_temporary_directory(self, net_fitted, mock_client):
        for call_args in mock_client.log_artifact.call_args_list:
            assert str(call_args[0][1]).startswith('/tmp')

    def test_dont_create_artifact(self, net_builder_cls, logger_mock_cls):
        logger = logger_mock_cls(create_artifact=False)
        net_builder_cls(callbacks=[logger], max_epochs=3)
        assert not logger.client_.log_artifact.called

    def test_run_terminated_automatically(self, net_fitted, mock_client):
        assert mock_client.set_terminated.call_count == 1

    def test_run_not_closed(self, logger_mock_cls, mock_client, net_builder_cls):
        logger = logger_mock_cls(terminate_after_train=False)
        net_builder_cls(callbacks=[logger], max_epochs=2)
        assert logger.client_.set_terminated.call_count == 0

    def test_fit_with_real_run_and_client(self, tmp_path, logger_cls, net_builder_cls):
        from mlflow.tracking import MlflowClient
        client = MlflowClient(tracking_uri=tmp_path.as_uri())
        experiment_name = 'foo'
        experiment_id = client.create_experiment(experiment_name)
        run = client.create_run(experiment_id)
        logger = logger_cls(run, client, create_artifact=False)
        net_builder_cls(callbacks=[logger], max_epochs=3)
        assert os.listdir(tmp_path)



"""Tests for virtual parameter setters"""
from unittest.mock import Mock

import pytest


class TestOptimizerSetter:

    @pytest.fixture
    def net_dummy(self):
        from skorch import NeuralNet
        net = Mock(spec=NeuralNet)
        net.lr = 0.01
        return net

    @pytest.fixture
    def optimizer_dummy(self):
        from torch.optim import Optimizer
        optim = Mock(spec=Optimizer)
        optim.param_groups = [
            {'lr': 0.01, 'momentum': 0.9},
            {'lr': 0.02, 'momentum': 0.9}
        ]
        return optim

    @pytest.fixture(scope='function')
    def net_optim_dummy(self, net_dummy, optimizer_dummy):
        net_dummy.optimizer_ = optimizer_dummy
        return net_dummy

    @pytest.fixture
    def setter(self):
        from skorch.setter import optimizer_setter
        return optimizer_setter

    def test_lr_attribute_is_updated(self, setter, net_optim_dummy):
        new_lr = net_optim_dummy.lr + 1
        setter(net_optim_dummy, 'lr', new_lr)

        assert net_optim_dummy.lr == new_lr

    def test_wrong_name_raises(self, setter, net_optim_dummy):
        # should be 'param_groups' instead
        param = 'optimizer__param_group__0__lr'
        value = 0.1
        with pytest.raises(AttributeError) as e:
            setter(net_optim_dummy, param, value)

        assert e.value.args[0] == (
            'Invalid parameter "{param}" for optimizer "optimizer"'
            .format(param=param)
        )

    @pytest.mark.parametrize('group', [0, 1])
    @pytest.mark.parametrize('sub_param, value', [
        ('momentum', 0.1),
        ('lr', 0.3),
    ])
    def test_only_specific_param_group_updated(self, setter, net_optim_dummy,
                                               group, sub_param, value):
        pgroups = net_optim_dummy.optimizer_.param_groups
        param = 'optimizer__param_groups__{}__{}'.format(group, sub_param)

        updated_group_pre = [g for i, g in enumerate(pgroups) if i == group]
        static_groups_pre = [g for i, g, in enumerate(pgroups) if i != group]
        assert len(updated_group_pre) == 1

        setter(net_optim_dummy, param, value)

        updated_group_new = [g for i, g in enumerate(pgroups) if i == group]
        static_groups_new = [g for i, g, in enumerate(pgroups) if i != group]

        assert updated_group_new[0][sub_param] == value
        assert all(old[sub_param] == new[sub_param] for old, new in zip(
            static_groups_pre, static_groups_new))



"""This script trains a NeuralNetClassifier on MNIST data, once with
skorch, once with pure PyTorch.

Apart from that change, both approaches are as close to eachother as
possible (e.g. performing the same validation steps).

At the end, we assert that the test accuracies are very close (1%
difference) and that skorch is not too much slower (at worst 30%
slower).

Call like this:

```
python examples/benchmarks/mnist.py
python examples/benchmarks/mnist.py --device cpu --num_samples 5000
```

When called the first time, this will download MNIST data to
examples/datasets/mldata (53M).

This script uses similar parameters as this one:
https://github.com/keras-team/keras/blob/0de2adf04b37aa972c955e69caf6917372b70a5b/examples/mnist_cnn.py


"""

import argparse
import time

import numpy as np
from sklearn.datasets import fetch_openml
from sklearn.model_selection import train_test_split
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import accuracy_score
from sklearn.utils import shuffle
import torch
from torch import nn

from skorch.utils import to_device
from skorch import NeuralNetClassifier
from skorch.callbacks import EpochScoring


BATCH_SIZE = 128
LEARNING_RATE = 0.1
MAX_EPOCHS = 12


def get_data(num_samples):
    mnist = fetch_openml('mnist_784', as_frame=False)
    torch.manual_seed(0)
    X = mnist.data.astype('float32').reshape(-1, 1, 28, 28)
    y = mnist.target.astype('int64')
    X, y = shuffle(X, y, random_state=0)
    X, y = X[:num_samples], y[:num_samples]
    X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=0)
    X_train /= 255
    X_test /= 255
    return X_train, X_test, y_train, y_test


class ClassifierModule(nn.Module):
    def __init__(self):
        super().__init__()

        self.cnn = nn.Sequential(
            nn.Conv2d(1, 32, (3, 3)),
            nn.ReLU(),
            nn.Conv2d(32, 64, (3, 3)),
            nn.ReLU(),
            nn.MaxPool2d((2, 2)),
            nn.Dropout(0.25),
        )
        self.out = nn.Sequential(
            nn.Linear(64 * 12 * 12, 128),
            nn.ReLU(),
            nn.Dropout(0.5),
            nn.Linear(128, 10),
            nn.Softmax(dim=-1),
        )

    def forward(self, X, **kwargs):
        X = self.cnn(X)
        X = X.reshape(-1, 64 * 12 * 12)
        X = self.out(X)
        return X


def performance_skorch(
        X_train,
        X_test,
        y_train,
        y_test,
        batch_size,
        device,
        lr,
        max_epochs,
):
    torch.manual_seed(0)
    net = NeuralNetClassifier(
        ClassifierModule,
        batch_size=batch_size,
        optimizer=torch.optim.Adadelta,
        lr=lr,
        device=device,
        max_epochs=max_epochs,
        callbacks=[
            ('tr_acc', EpochScoring(
                'accuracy',
                lower_is_better=False,
                on_train=True,
                name='train_acc',
            )),
        ],
    )
    net.fit(X_train, y_train)
    y_pred = net.predict(X_test)
    score = accuracy_score(y_test, y_pred)
    return score


def report(losses, batch_sizes, y, y_proba, epoch, time, training=True):
    template = "{} | epoch {:>2} | "

    loss = np.average(losses, weights=batch_sizes)
    y_pred = np.argmax(y_proba, axis=1)
    acc = accuracy_score(y, y_pred)

    template += "acc: {:.4f} | loss: {:.4f} | time: {:.2f}"
    print(template.format(
        'train' if training else 'valid', epoch + 1, acc, loss, time))


def train_torch(
        model,
        X,
        y,
        batch_size,
        device,
        lr,
        max_epochs,
):
    model = to_device(model, device)

    idx_train, idx_valid = next(iter(StratifiedKFold(5).split(np.arange(len(X)), y)))
    X_train, X_valid, y_train, y_valid = (
        X[idx_train], X[idx_valid], y[idx_train], y[idx_valid])
    dataset_train = torch.utils.data.TensorDataset(
        torch.tensor(X_train),
        torch.tensor(y_train),
    )
    iterator_train = torch.utils.data.DataLoader(dataset_train, batch_size=batch_size)
    dataset_valid = torch.utils.data.TensorDataset(
        torch.tensor(X_valid),
        torch.tensor(y_valid),
    )
    iterator_valid = torch.utils.data.DataLoader(dataset_valid, batch_size=batch_size)

    optimizer = torch.optim.Adadelta(model.parameters(), lr=lr)
    criterion = nn.NLLLoss()

    for epoch in range(max_epochs):
        train_out = train_step(
            model,
            iterator_train,
            batch_size=batch_size,
            device=device,
            criterion=criterion,
            optimizer=optimizer,
        )
        report(y=y_train, epoch=epoch, training=True, **train_out)

        valid_out = valid_step(
            model,
            iterator_valid,
            batch_size=batch_size,
            device=device,
            criterion=criterion,
        )
        report(y=y_valid, epoch=epoch, training=False, **valid_out)

        print('-' * 50)

    return model


def train_step(model, iterator, device, criterion, batch_size, optimizer):
    model.train()
    y_preds = []
    losses = []
    batch_sizes = []
    tic = time.time()
    for Xi, yi in iterator:
        Xi, yi = to_device(Xi, device), to_device(yi, device)
        optimizer.zero_grad()
        y_pred = model(Xi)
        y_pred = torch.log(y_pred)
        loss = criterion(y_pred, yi)
        loss.backward()
        optimizer.step()

        y_preds.append(y_pred)
        losses.append(loss.item())
        batch_sizes.append(len(Xi))
    toc = time.time()
    return {
        'losses': losses,
        'batch_sizes': batch_sizes,
        'y_proba': torch.cat(y_preds).cpu().detach().numpy(),
        'time': toc - tic,
    }


def valid_step(model, iterator, device, criterion, batch_size):
    model.eval()
    y_preds = []
    losses = []
    batch_sizes = []
    tic = time.time()
    with torch.no_grad():
        for Xi, yi in iterator:
            Xi, yi = to_device(Xi, device), to_device(yi, device)
            y_pred = model(Xi)
            y_pred = torch.log(y_pred)
            loss = criterion(y_pred, yi)

            y_preds.append(y_pred)
            loss = loss.item()
            losses.append(loss)
            batch_sizes.append(len(Xi))
    toc = time.time()
    return {
        'losses': losses,
        'batch_sizes': batch_sizes,
        'y_proba': torch.cat(y_preds).cpu().detach().numpy(),
        'time': toc - tic,
    }


def performance_torch(
        X_train,
        X_test,
        y_train,
        y_test,
        batch_size,
        device,
        lr,
        max_epochs,
):
    torch.manual_seed(0)
    model = ClassifierModule()
    model = train_torch(
        model,
        X_train,
        y_train,
        batch_size=batch_size,
        device=device,
        max_epochs=max_epochs,
        lr=lr,
    )

    X_test = torch.tensor(X_test).to(device)
    with torch.no_grad():
        y_pred = model(X_test).cpu().numpy().argmax(1)
    return accuracy_score(y_test, y_pred)


def main(device, num_samples):
    data = get_data(num_samples)
    # trigger potential cuda call overhead
    torch.zeros(1).to(device)

    print("\nTesting skorch performance")
    tic = time.time()
    score_skorch = performance_skorch(
        *data,
        batch_size=BATCH_SIZE,
        max_epochs=MAX_EPOCHS,
        lr=LEARNING_RATE,
        device=device,
    )
    time_skorch = time.time() - tic

    print("\nTesting pure torch performance")
    tic = time.time()
    score_torch = performance_torch(
        *data,
        batch_size=BATCH_SIZE,
        max_epochs=MAX_EPOCHS,
        lr=LEARNING_RATE,
        device=device,
    )
    time_torch = time.time() - tic

    print("time skorch: {:.4f}, time torch: {:.4f}".format(
        time_skorch, time_torch))
    print("score skorch: {:.4f}, score torch: {:.4f}".format(
        score_skorch, score_torch))

    assert np.isclose(score_skorch, score_torch, rtol=0.01), "Scores are not close enough."
    assert np.isclose(time_skorch, time_torch, rtol=0.3), "Times are not close enough."


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="skorch MNIST benchmark")
    parser.add_argument('--device', type=str, default='cuda',
                        help='device (e.g. "cuda", "cpu")')
    parser.add_argument('--num_samples', type=int, default=20000,
                        help='total number of samples to use')
    args = parser.parse_args()
    main(device=args.device, num_samples=args.num_samples)



"""Contains skorch-specific exceptions and warnings."""

from sklearn.exceptions import NotFittedError


class SkorchException(BaseException):
    """Base skorch exception."""


class NotInitializedError(SkorchException, NotFittedError):
    """Module is not initialized, please call the ``.initialize``
    method or train the model by calling ``.fit(...)``.

    """


class SkorchAttributeError(SkorchException):
    """An attribute was set incorrectly on a skorch net."""


class SkorchWarning(UserWarning):
    """Base skorch warning."""


class DeviceWarning(SkorchWarning):
    """A problem with a device (e.g. CUDA) was detected."""


class SkorchTrainingImpossibleError(SkorchException):
    """The net cannot be used for training"""
